"""
Tests common to psycopg.AsyncCursor and its subclasses.
"""

import weakref
import datetime as dt
from typing import Any
from contextlib import aclosing

import pytest
from packaging.version import parse as ver

import psycopg
from psycopg import pq, rows, sql
from psycopg.adapt import PyFormat
from psycopg.types import TypeInfo

from . import _test_cursor
from .utils import raiseif
from .acompat import alist, gather, spawn
from .fix_crdb import crdb_encoding
from .test_adapt import make_loader
from ._test_cursor import my_row_factory, ph

execmany = _test_cursor.execmany  # avoid F811 underneath
_execmany = _test_cursor._execmany  # needed by the execmany fixture


cursor_classes = [psycopg.AsyncCursor, psycopg.AsyncClientCursor]
# Allow to import (not necessarily to run) the module with psycopg 3.1.
# Needed to test psycopg_pool 3.2 tests with psycopg 3.1 imported, i.e. to run
# `pytest -m pool`. (which might happen when releasing pool packages).
if ver(psycopg.__version__) >= ver("3.2.0.dev0"):
    cursor_classes.append(psycopg.AsyncRawCursor)


@pytest.fixture(params=cursor_classes)
async def aconn(aconn, request, anyio_backend):
    aconn.cursor_factory = request.param
    return aconn


async def test_init(aconn):
    cur = aconn.cursor_factory(aconn)
    await cur.execute("select 1")
    assert (await cur.fetchone()) == (1,)

    aconn.row_factory = rows.dict_row
    cur = aconn.cursor_factory(aconn)
    await cur.execute("select 1 as a")
    assert (await cur.fetchone()) == {"a": 1}


async def test_init_factory(aconn):
    cur = aconn.cursor_factory(aconn, row_factory=rows.dict_row)
    await cur.execute("select 1 as a")
    assert (await cur.fetchone()) == {"a": 1}


async def test_close(aconn):
    cur = aconn.cursor()
    assert not cur.closed
    await cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        await cur.execute("select 'foo'")

    await cur.close()
    assert cur.closed


async def test_cursor_close_fetchone(aconn):
    cur = aconn.cursor()
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    await cur.execute(query)
    for _ in range(5):
        await cur.fetchone()

    await cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        await cur.fetchone()


async def test_cursor_close_fetchmany(aconn):
    cur = aconn.cursor()
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    await cur.execute(query)
    assert len(await cur.fetchmany(2)) == 2

    await cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        await cur.fetchmany(2)


async def test_cursor_close_fetchall(aconn):
    cur = aconn.cursor()
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    await cur.execute(query)
    assert len(await cur.fetchall()) == 10

    await cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        await cur.fetchall()


async def test_context(aconn):
    async with aconn.cursor() as cur:
        assert not cur.closed

    assert cur.closed


@pytest.mark.slow
async def test_weakref(aconn, gc_collect):
    cur = aconn.cursor()
    w = weakref.ref(cur)
    await cur.close()
    del cur
    gc_collect()
    assert w() is None


async def test_pgresult(aconn):
    cur = aconn.cursor()
    await cur.execute("select 1")
    assert cur.pgresult
    await cur.close()
    assert not cur.pgresult


async def test_statusmessage(aconn):
    cur = aconn.cursor()
    assert cur.statusmessage is None

    await cur.execute("select generate_series(1, 10)")
    assert cur.statusmessage == "SELECT 10"

    await cur.execute("")
    assert cur.statusmessage is None

    await cur.execute("create table statusmessage ()")
    assert cur.statusmessage == "CREATE TABLE"

    with pytest.raises(psycopg.ProgrammingError):
        await cur.execute("wat")
    assert cur.statusmessage is None


async def test_execute_sql(aconn):
    cur = aconn.cursor()
    await cur.execute(sql.SQL("select {value}").format(value="hello"))
    assert (await cur.fetchone()) == ("hello",)


async def test_next(aconn):
    cur = aconn.cursor()
    await cur.execute("select 1")
    assert await anext(cur) == (1,)
    with pytest.raises(StopAsyncIteration):
        await anext(cur)


async def test_query_parse_cache_size(aconn):
    cur = aconn.cursor()
    cls = type(cur)

    # Warning: testing internal structures. Test might need refactoring with the code.
    cache: Any
    if cls is psycopg.AsyncCursor:
        cache = psycopg._queries._query2pg
    elif cls is psycopg.AsyncClientCursor:
        cache = psycopg._queries._query2pg_client
    elif cls is psycopg.AsyncRawCursor:
        pytest.skip("RawCursor has no query parse cache")
    else:
        assert False, cls

    cache.cache_clear()
    ci = cache.cache_info()
    h0, m0 = ci.hits, ci.misses
    tests = [
        (f"select 1 -- {'x' * 3500}", (), h0, m0 + 1),
        (f"select 1 -- {'x' * 3500}", (), h0 + 1, m0 + 1),
        (f"select 1 -- {'x' * 4500}", (), h0 + 1, m0 + 1),
        (f"select 1 -- {'x' * 4500}", (), h0 + 1, m0 + 1),
        (f"select 1 -- {'%s' * 40}", ("x",) * 40, h0 + 1, m0 + 2),
        (f"select 1 -- {'%s' * 40}", ("x",) * 40, h0 + 2, m0 + 2),
        (f"select 1 -- {'%s' * 60}", ("x",) * 60, h0 + 2, m0 + 2),
        (f"select 1 -- {'%s' * 60}", ("x",) * 60, h0 + 2, m0 + 2),
    ]
    for i, (query, params, hits, misses) in enumerate(tests):
        pq = cur._query_cls(psycopg.adapt.Transformer())
        pq.convert(query, params)
        ci = cache.cache_info()
        assert ci.hits == hits, f"at {i}"
        assert ci.misses == misses, f"at {i}"


async def test_execute_many_results(aconn):
    cur = aconn.cursor()
    assert cur.nextset() is None

    rv = await cur.execute("select 'foo'; select generate_series(1,3)")
    assert rv is cur
    assert (await cur.fetchall()) == [("foo",)]
    assert cur.rowcount == 1
    assert cur.nextset()
    assert (await cur.fetchall()) == [(1,), (2,), (3,)]
    assert cur.rowcount == 3
    assert cur.nextset() is None

    await cur.close()
    assert cur.nextset() is None


async def test_set_results(aconn):
    cur = aconn.cursor()

    with pytest.raises(IndexError):
        await cur.set_result(0)

    await cur.execute("select 'foo'; select generate_series(1,3)")
    assert await cur.set_result(0) is cur
    assert (await cur.fetchall()) == [("foo",)]
    assert cur.rowcount == 1

    assert await cur.set_result(-1) is cur
    assert (await cur.fetchall()) == [(1,), (2,), (3,)]
    assert cur.rowcount == 3

    with pytest.raises(IndexError):
        await cur.set_result(2)

    with pytest.raises(IndexError):
        await cur.set_result(-3)


async def test_execute_sequence(aconn):
    cur = aconn.cursor()
    rv = await cur.execute(
        ph(cur, "select %s::int, %s::text, %s::text"), [1, "foo", None]
    )
    assert rv is cur
    assert len(cur._results) == 1
    assert cur.pgresult.get_value(0, 0) == b"1"
    assert cur.pgresult.get_value(0, 1) == b"foo"
    assert cur.pgresult.get_value(0, 2) is None
    assert cur.nextset() is None


@pytest.mark.parametrize("query", ["", " ", ";"])
async def test_execute_empty_query(aconn, query):
    cur = aconn.cursor()
    await cur.execute(query)
    assert cur.pgresult.status == pq.ExecStatus.EMPTY_QUERY
    with pytest.raises(psycopg.ProgrammingError):
        await cur.fetchone()


async def test_execute_type_change(aconn):
    # issue #112
    await aconn.execute("create table bug_112 (num integer)")
    cur = aconn.cursor()
    sql = ph(cur, "insert into bug_112 (num) values (%s)")
    await cur.execute(sql, (1,))
    await cur.execute(sql, (100_000,))
    await cur.execute("select num from bug_112 order by num")
    assert (await cur.fetchall()) == [(1,), (100_000,)]


async def test_executemany_type_change(aconn):
    await aconn.execute("create table bug_112 (num integer)")
    cur = aconn.cursor()
    sql = ph(cur, "insert into bug_112 (num) values (%s)")
    await cur.executemany(sql, [(1,), (100_000,)])
    await cur.execute("select num from bug_112 order by num")
    assert (await cur.fetchall()) == [(1,), (100_000,)]


@pytest.mark.parametrize(
    "query", ["copy testcopy from stdin", "copy testcopy to stdout"]
)
async def test_execute_copy(aconn, query):
    cur = aconn.cursor()
    await cur.execute("create table testcopy (id int)")
    with pytest.raises(psycopg.ProgrammingError):
        await cur.execute(query)


async def test_fetchone(aconn):
    cur = aconn.cursor()
    await cur.execute(ph(cur, "select %s::int, %s::text, %s::text"), [1, "foo", None])
    assert cur.pgresult.fformat(0) == 0

    row = await cur.fetchone()
    assert row == (1, "foo", None)
    row = await cur.fetchone()
    assert row is None


async def test_binary_cursor_execute(aconn):
    with raiseif(
        aconn.cursor_factory is psycopg.AsyncClientCursor, psycopg.NotSupportedError
    ) as ex:
        cur = aconn.cursor(binary=True)
        await cur.execute(ph(cur, "select %s, %s"), [1, None])
    if ex:
        return

    assert (await cur.fetchone()) == (1, None)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x01"


async def test_execute_binary(aconn):
    cur = aconn.cursor()
    with raiseif(
        aconn.cursor_factory is psycopg.AsyncClientCursor, psycopg.NotSupportedError
    ) as ex:
        await cur.execute(ph(cur, "select %s, %s"), [1, None], binary=True)
    if ex:
        return

    assert (await cur.fetchone()) == (1, None)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x01"


async def test_binary_cursor_text_override(aconn):
    cur = aconn.cursor(binary=True)
    await cur.execute(ph(cur, "select %s, %s"), [1, None], binary=False)
    assert (await cur.fetchone()) == (1, None)
    assert cur.pgresult.fformat(0) == 0
    assert cur.pgresult.get_value(0, 0) == b"1"


@pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
async def test_query_encode(aconn, encoding):
    await aconn.execute(f"set client_encoding to {encoding}")
    cur = aconn.cursor()
    await cur.execute("select '\u20ac'")
    (res,) = await cur.fetchone()
    assert res == "\u20ac"


@pytest.mark.parametrize("encoding", [crdb_encoding("latin1")])
async def test_query_badenc(aconn, encoding):
    await aconn.execute(f"set client_encoding to {encoding}")
    cur = aconn.cursor()
    with pytest.raises(UnicodeEncodeError):
        await cur.execute("select '\u20ac'")


async def test_executemany(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
    )
    await cur.execute("select num, data from execmany order by 1")
    rv = await cur.fetchall()
    assert rv == [(10, "hello"), (20, "world")]


async def test_executemany_name(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%(num)s, %(data)s)"),
        [{"num": 11, "data": "hello", "x": 1}, {"num": 21, "data": "world"}],
    )
    await cur.execute("select num, data from execmany order by 1")
    rv = await cur.fetchall()
    assert rv == [(11, "hello"), (21, "world")]


async def test_executemany_no_data(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"), []
    )
    assert cur.rowcount == 0


async def test_executemany_rowcount(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
    )
    assert cur.rowcount == 2


@pytest.mark.parametrize("returning", [False, True])
async def test_executemany_statusmessage(aconn, execmany, returning):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
        returning=returning,
    )
    assert cur.rowcount == (1 if returning else 2)
    assert cur.statusmessage is not None
    assert cur.statusmessage.startswith("INSERT")

    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [],
        returning=returning,
    )
    assert cur.statusmessage is None


async def test_executemany_returning(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s) returning num"),
        [(10, "hello"), (20, "world")],
        returning=True,
    )
    assert cur.rowcount == 1
    assert (await cur.fetchone()) == (10,)
    assert cur.nextset()
    assert cur.rowcount == 1
    assert (await cur.fetchone()) == (20,)
    assert cur.nextset() is None


async def test_executemany_returning_discard(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s) returning num"),
        [(10, "hello"), (20, "world")],
    )
    assert cur.rowcount == 2
    with pytest.raises(psycopg.ProgrammingError):
        await cur.fetchone()
    assert cur.nextset() is None


async def test_executemany_no_result(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
        returning=True,
    )
    assert cur.rowcount == 1
    assert cur.statusmessage.startswith("INSERT")
    with pytest.raises(psycopg.ProgrammingError):
        await cur.fetchone()
    pgresult = cur.pgresult
    assert cur.nextset()
    assert cur.rowcount == 1
    assert cur.statusmessage.startswith("INSERT")
    assert pgresult is not cur.pgresult
    assert cur.nextset() is None


async def test_executemany_rowcount_no_hit(aconn, execmany):
    cur = aconn.cursor()
    await cur.executemany(ph(cur, "delete from execmany where id = %s"), [(-1,), (-2,)])
    assert cur.rowcount == 0
    await cur.executemany(ph(cur, "delete from execmany where id = %s"), [])
    assert cur.rowcount == 0
    await cur.executemany(
        ph(cur, "delete from execmany where id = %s returning num"), [(-1,), (-2,)]
    )
    assert cur.rowcount == 0


@pytest.mark.parametrize(
    "query",
    [
        "insert into nosuchtable values (%s, %s)",
        "copy (select %s, %s) to stdout",
        "wat (%s, %s)",
    ],
)
async def test_executemany_badquery(aconn, query):
    cur = aconn.cursor()
    with pytest.raises(psycopg.DatabaseError):
        await cur.executemany(ph(cur, query), [(10, "hello"), (20, "world")])


@pytest.mark.parametrize("fmt_in", PyFormat)
async def test_executemany_null_first(aconn, fmt_in):
    cur = aconn.cursor()
    await cur.execute("create table testmany (a bigint, b bigint)")
    await cur.executemany(
        ph(cur, f"insert into testmany values (%{fmt_in.value}, %{fmt_in.value})"),
        [[1, None], [3, 4]],
    )
    with pytest.raises((psycopg.DataError, psycopg.ProgrammingError)):
        await cur.executemany(
            ph(cur, f"insert into testmany values (%{fmt_in.value}, %{fmt_in.value})"),
            [[1, ""], [3, 4]],
        )


@pytest.mark.slow
async def test_executemany_lock(aconn):
    async def do_execmany():
        async with aconn.cursor() as cur:
            await cur.executemany(
                ph(cur, "select pg_sleep(%s)"), [(0.1,) for _ in range(10)]
            )

    async def do_exec():
        async with aconn.cursor() as cur:
            for i in range(100):
                await cur.execute("select 1")

    await gather(spawn(do_execmany), spawn(do_exec))


async def test_rowcount(aconn):
    cur = aconn.cursor()

    await cur.execute("select 1 from generate_series(1, 0)")
    assert cur.rowcount == 0

    await cur.execute("select 1 from generate_series(1, 42)")
    assert cur.rowcount == 42

    await cur.execute("show timezone")
    assert cur.rowcount == 1

    await cur.execute("create table test_rowcount_notuples (id int primary key)")
    assert cur.rowcount == -1

    await cur.execute(
        "insert into test_rowcount_notuples select generate_series(1, 42)"
    )
    assert cur.rowcount == 42


async def test_rownumber(aconn):
    cur = aconn.cursor()
    assert cur.rownumber is None

    await cur.execute("select 1 from generate_series(1, 42)")
    assert cur.rownumber == 0

    await cur.fetchone()
    assert cur.rownumber == 1
    await cur.fetchone()
    assert cur.rownumber == 2
    await cur.fetchmany(10)
    assert cur.rownumber == 12
    rns: list[int] = []
    async for i in cur:
        assert cur.rownumber
        rns.append(cur.rownumber)
        if len(rns) >= 3:
            break
    assert rns == [13, 14, 15]
    assert len(await cur.fetchall()) == 42 - rns[-1]
    assert cur.rownumber == 42


@pytest.mark.parametrize("query", ["", "set timezone to utc"])
async def test_rownumber_none(aconn, query):
    cur = aconn.cursor()
    await cur.execute(query)
    assert cur.rownumber is None


async def test_rownumber_mixed(aconn):
    cur = aconn.cursor()
    await cur.execute("""
select x from generate_series(1, 3) x;
set timezone to utc;
select x from generate_series(4, 6) x;
""")
    assert cur.rownumber == 0
    assert await cur.fetchone() == (1,)
    assert cur.rownumber == 1
    assert await cur.fetchone() == (2,)
    assert cur.rownumber == 2
    cur.nextset()
    assert cur.rownumber is None
    cur.nextset()
    assert cur.rownumber == 0
    assert await cur.fetchone() == (4,)
    assert cur.rownumber == 1


async def test_iter(aconn):
    cur = aconn.cursor()
    await cur.execute("select generate_series(1, 3)")
    assert await alist(cur) == [(1,), (2,), (3,)]


async def test_iter_stop(aconn):
    cur = aconn.cursor()
    await cur.execute("select generate_series(1, 3)")
    async for rec in cur:
        assert rec == (1,)
        break

    async for rec in cur:
        assert rec == (2,)
        break

    assert (await cur.fetchone()) == (3,)
    assert (await alist(cur)) == []


async def test_row_factory(aconn):
    cur = aconn.cursor(row_factory=my_row_factory)

    await cur.execute("reset search_path")
    with pytest.raises(psycopg.ProgrammingError):
        await cur.fetchone()

    await cur.execute("select 'foo' as bar")
    (r,) = await cur.fetchone()
    assert r == "FOObar"

    await cur.execute("select 'x' as x; select 'y' as y, 'z' as z")
    assert await cur.fetchall() == [["Xx"]]
    assert cur.nextset()
    assert await cur.fetchall() == [["Yy", "Zz"]]

    await cur.scroll(-1)
    cur.row_factory = rows.dict_row
    assert await cur.fetchone() == {"y": "y", "z": "z"}


async def test_row_factory_none(aconn):
    cur = aconn.cursor(row_factory=None)
    assert cur.row_factory is rows.tuple_row
    await cur.execute("select 1 as a, 2 as b")
    r = await cur.fetchone()
    assert type(r) is tuple
    assert r == (1, 2)


async def test_bad_row_factory(aconn):
    def broken_factory(cur):
        1 / 0

    cur = aconn.cursor(row_factory=broken_factory)
    with pytest.raises(ZeroDivisionError):
        await cur.execute("select 1")

    def broken_maker(cur):
        def make_row(seq):
            1 / 0

        return make_row

    cur = aconn.cursor(row_factory=broken_maker)
    await cur.execute("select 1")
    with pytest.raises(ZeroDivisionError):
        await cur.fetchone()


async def test_scroll(aconn):
    cur = aconn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        await cur.scroll(0)

    await cur.execute("select generate_series(0,9)")
    await cur.scroll(2)
    assert await cur.fetchone() == (2,)
    await cur.scroll(2)
    assert await cur.fetchone() == (5,)
    await cur.scroll(2, mode="relative")
    assert await cur.fetchone() == (8,)
    await cur.scroll(-1)
    assert await cur.fetchone() == (8,)
    await cur.scroll(-2)
    assert await cur.fetchone() == (7,)
    await cur.scroll(2, mode="absolute")
    assert await cur.fetchone() == (2,)

    # on the boundary
    await cur.scroll(0, mode="absolute")
    assert await cur.fetchone() == (0,)
    with pytest.raises(IndexError):
        await cur.scroll(-1, mode="absolute")

    await cur.scroll(0, mode="absolute")
    with pytest.raises(IndexError):
        await cur.scroll(-1)

    await cur.scroll(9, mode="absolute")
    assert await cur.fetchone() == (9,)
    with pytest.raises(IndexError):
        await cur.scroll(10, mode="absolute")

    await cur.scroll(9, mode="absolute")
    with pytest.raises(IndexError):
        await cur.scroll(1)

    with pytest.raises(ValueError):
        await cur.scroll(1, "wat")


@pytest.mark.parametrize(
    "query, params, want",
    [
        ("select %(x)s", {"x": 1}, (1,)),
        ("select %(x)s, %(y)s", {"x": 1, "y": 2}, (1, 2)),
        ("select %(x)s, %(x)s", {"x": 1}, (1, 1)),
    ],
)
async def test_execute_params_named(aconn, query, params, want):
    cur = aconn.cursor()
    await cur.execute(ph(cur, query), params)
    rec = await cur.fetchone()
    assert rec == want


async def test_stream(aconn):
    cur = aconn.cursor()
    recs = []
    async for rec in cur.stream(
        ph(cur, "select i, '2021-01-01'::date + i from generate_series(1, %s) as i"),
        [2],
    ):
        recs.append(rec)

    assert recs == [(1, dt.date(2021, 1, 2)), (2, dt.date(2021, 1, 3))]


async def test_stream_sql(aconn):
    cur = aconn.cursor()
    recs = await alist(
        cur.stream(
            sql.SQL(
                "select i, '2021-01-01'::date + i from generate_series(1, {}) as i"
            ).format(2)
        )
    )

    assert recs == [(1, dt.date(2021, 1, 2)), (2, dt.date(2021, 1, 3))]


async def test_stream_row_factory(aconn):
    cur = aconn.cursor(row_factory=rows.dict_row)
    it = cur.stream("select generate_series(1,2) as a")
    assert (await anext(it))["a"] == 1
    cur.row_factory = rows.namedtuple_row
    assert (await anext(it)).a == 2


async def test_stream_no_row(aconn):
    cur = aconn.cursor()
    recs = await alist(cur.stream("select generate_series(2,1) as a"))
    assert recs == []


async def test_stream_chunked_invalid_size(aconn):
    cur = aconn.cursor()
    with pytest.raises(ValueError, match=r"size must be >= 1"):
        await anext(cur.stream("select 1", size=0))


@pytest.mark.libpq("< 17")
async def test_stream_chunked_not_supported(aconn):
    cur = aconn.cursor()
    with pytest.raises(psycopg.NotSupportedError):
        await anext(cur.stream("select generate_series(1, 4)", size=2))


@pytest.mark.libpq(">= 17")
async def test_stream_chunked(aconn):
    cur = aconn.cursor()
    recs = await alist(cur.stream("select generate_series(1, 5) as a", size=2))
    assert recs == [(1,), (2,), (3,), (4,), (5,)]


@pytest.mark.libpq(">= 17")
async def test_stream_chunked_row_factory(aconn):
    cur = aconn.cursor(row_factory=rows.scalar_row)
    it = cur.stream("select generate_series(1, 5) as a", size=2)
    for i in range(1, 6):
        assert await anext(it) == i
        assert [c.name for c in cur.description] == ["a"]


@pytest.mark.crdb_skip("no col query")
async def test_stream_no_col(aconn):
    cur = aconn.cursor()
    recs = await alist(cur.stream("select"))
    assert recs == [()]


@pytest.mark.parametrize(
    "query",
    [
        "create table test_stream_badq ()",
        "copy (select 1) to stdout",
        "wat?",
    ],
)
async def test_stream_badquery(aconn, query):
    cur = aconn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        async for rec in cur.stream(query):
            pass


async def test_stream_error_tx(aconn):
    cur = aconn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        async for rec in cur.stream("wat"):
            pass
    assert aconn.info.transaction_status == pq.TransactionStatus.INERROR


async def test_stream_error_notx(aconn):
    await aconn.set_autocommit(True)
    cur = aconn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        async for rec in cur.stream("wat"):
            pass
    assert aconn.info.transaction_status == pq.TransactionStatus.IDLE


async def test_stream_error_python_to_consume(aconn):
    cur = aconn.cursor()
    with pytest.raises(ZeroDivisionError):
        async with aclosing(cur.stream("select generate_series(1, 10000)")) as gen:
            async for rec in gen:
                1 / 0
    assert aconn.info.transaction_status in (
        pq.TransactionStatus.INTRANS,
        pq.TransactionStatus.INERROR,
    )


async def test_stream_error_python_consumed(aconn):
    cur = aconn.cursor()
    with pytest.raises(ZeroDivisionError):
        gen = cur.stream("select 1")
        async for rec in gen:
            1 / 0

    await gen.aclose()
    assert aconn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("autocommit", [False, True])
async def test_stream_close(aconn, autocommit):
    await aconn.set_autocommit(autocommit)
    cur = aconn.cursor()
    with pytest.raises(psycopg.OperationalError):
        async for rec in cur.stream("select generate_series(1, 3)"):
            if rec[0] == 1:
                await aconn.close()
            else:
                assert False

    assert aconn.closed


async def test_stream_binary_cursor(aconn):
    with raiseif(
        aconn.cursor_factory is psycopg.AsyncClientCursor, psycopg.NotSupportedError
    ):
        cur = aconn.cursor(binary=True)
        recs = []
        async for rec in cur.stream("select x::int4 from generate_series(1, 2) x"):
            recs.append(rec)
            assert cur.pgresult.fformat(0) == 1
            assert cur.pgresult.get_value(0, 0) == bytes([0, 0, 0, rec[0]])

        assert recs == [(1,), (2,)]


async def test_stream_execute_binary(aconn):
    cur = aconn.cursor()
    recs = []
    with raiseif(
        aconn.cursor_factory is psycopg.AsyncClientCursor, psycopg.NotSupportedError
    ):
        async for rec in cur.stream(
            "select x::int4 from generate_series(1, 2) x", binary=True
        ):
            recs.append(rec)
            assert cur.pgresult.fformat(0) == 1
            assert cur.pgresult.get_value(0, 0) == bytes([0, 0, 0, rec[0]])

        assert recs == [(1,), (2,)]


async def test_stream_binary_cursor_text_override(aconn):
    cur = aconn.cursor(binary=True)
    recs = []
    async for rec in cur.stream("select generate_series(1, 2)", binary=False):
        recs.append(rec)
        assert cur.pgresult.fformat(0) == 0
        assert cur.pgresult.get_value(0, 0) == str(rec[0]).encode()

    assert recs == [(1,), (2,)]


async def test_str(aconn):
    cur = aconn.cursor()
    assert "[IDLE]" in str(cur)
    assert "[closed]" not in str(cur)
    assert "[no result]" in str(cur)
    await cur.execute("select 1")
    assert "[INTRANS]" in str(cur)
    assert "[TUPLES_OK]" in str(cur)
    assert "[closed]" not in str(cur)
    assert "[no result]" not in str(cur)
    await cur.close()
    assert "[closed]" in str(cur)
    assert "[INTRANS]" in str(cur)


@pytest.mark.pipeline
async def test_message_0x33(aconn):
    # https://github.com/psycopg/psycopg/issues/314
    notices = []
    aconn.add_notice_handler(lambda diag: notices.append(diag.message_primary))

    await aconn.set_autocommit(True)
    async with aconn.pipeline():
        cur = await aconn.execute("select 'test'")
        assert (await cur.fetchone()) == ("test",)

    assert not notices


async def test_typeinfo(aconn):
    info = await TypeInfo.fetch(aconn, "jsonb")
    assert info is not None


async def test_error_no_result(aconn):
    cur = aconn.cursor()
    with pytest.raises(psycopg.ProgrammingError, match="no result available"):
        await cur.fetchone()

    await cur.execute("set timezone to utc")
    with pytest.raises(
        psycopg.ProgrammingError, match="last operation.*command status: SET"
    ):
        await cur.fetchone()

    await cur.execute("")
    with pytest.raises(
        psycopg.ProgrammingError, match="last operation.*result status: EMPTY_QUERY"
    ):
        await cur.fetchone()


async def test_row_maker_returns_none(aconn):
    cur = aconn.cursor(row_factory=rows.scalar_row)
    query = "values (null), (0)"
    recs = [None, 0]

    await cur.execute(query)
    assert [await cur.fetchone() for _ in range(len(recs))] == recs
    await cur.execute(query)
    assert await cur.fetchmany(len(recs)) == recs
    await cur.execute(query)
    assert await cur.fetchall() == recs
    await cur.execute(query)
    assert await alist(cur) == recs
    stream = cur.stream(query)
    assert await alist(stream) == recs


@pytest.mark.parametrize("count", [1, 3])
async def test_results_after_execute(aconn, count):
    async with aconn.cursor() as cur:
        await cur.execute(
            ";".join(f"select * from generate_series(1, {i})" for i in range(count))
        )
        ress = await alist(await res.fetchall() async for res in cur.results())
        assert ress == [[(j + 1,) for j in range(i)] for i in range(count)]


@pytest.mark.parametrize("count", [0, 1, 3])
@pytest.mark.parametrize("returning", [False, True])
async def test_results_after_executemany(aconn, count, returning):
    async with aconn.cursor() as cur:
        await cur.executemany(
            ph(cur, "select * from generate_series(1, %s)"),
            [(i,) for i in range(count)],
            returning=returning,
        )
        ress = await alist(await res.fetchall() async for res in cur.results())
        if returning:
            assert ress == [[(j + 1,) for j in range(i)] for i in range(count)]
        else:
            assert ress == []


async def test_change_loader_results(aconn):
    cur = aconn.cursor()
    # With no result
    cur.adapters.register_loader("text", make_loader("1"))

    await cur.execute("""
        values ('foo'::text);
        values ('bar'::text), ('baz');
        values ('qux'::text);
        """)
    assert (await cur.fetchall()) == [("foo1",)]

    cur.nextset()
    assert (await cur.fetchone()) == ("bar1",)
    cur.adapters.register_loader("text", make_loader("2"))
    assert (await cur.fetchone()) == ("baz2",)
    await cur.scroll(-2)
    assert (await cur.fetchall()) == [("bar2",), ("baz2",)]

    cur.nextset()
    assert (await cur.fetchall()) == [("qux2",)]

    # After the final result
    assert not cur.nextset()
    cur.adapters.register_loader("text", make_loader("3"))
    assert (await cur.fetchone()) is None
    await cur.set_result(0)
    assert (await cur.fetchall()) == [("foo3",)]



"""
Psycopg AsyncPipeline object implementation.
"""

# Copyright (C) 2021 The Psycopg Team

from __future__ import annotations

import logging
from types import TracebackType
from typing import TYPE_CHECKING, Any

from . import errors as e
from ._compat import Self
from ._pipeline_base import BasePipeline

if TYPE_CHECKING:
    from .connection_async import AsyncConnection

logger = logging.getLogger("psycopg")


class _DummyLock:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass


class AsyncPipeline(BasePipeline):
    """Handler for (async) connection in pipeline mode."""

    __module__ = "psycopg"
    _conn: AsyncConnection[Any]

    def __init__(self, conn: AsyncConnection[Any], _no_lock: bool = False) -> None:
        super().__init__(conn)
        self._lock = _DummyLock() if _no_lock else conn.lock

    async def sync(self) -> None:
        """Sync the pipeline, send any pending command and receive and process
        all available results.
        """
        try:
            async with self._lock:
                await self._conn.wait(self._sync_gen())
        except e._NO_TRACEBACK as ex:
            raise ex.with_traceback(None)

    async def __aenter__(self) -> Self:
        async with self._lock:
            await self._conn.wait(self._enter_gen())
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        try:
            async with self._lock:
                await self._conn.wait(self._exit_gen())
        except Exception as exc2:
            # Don't clobber an exception raised in the block with this one
            if exc_val:
                logger.warning("error ignored terminating %r: %s", self, exc2)
            else:
                raise exc2.with_traceback(None)
        finally:
            self._exit(exc_val)



import pytest

from ..test_typing import _test_reveal


@pytest.mark.parametrize(
    "conn, type",
    [
        (
            "psycopg.crdb.connect()",
            "psycopg.crdb.CrdbConnection[Tuple[Any, ...]]",
        ),
        (
            "psycopg.crdb.connect(row_factory=rows.dict_row)",
            "psycopg.crdb.CrdbConnection[Dict[str, Any]]",
        ),
        (
            "psycopg.crdb.CrdbConnection.connect()",
            "psycopg.crdb.CrdbConnection[Tuple[Any, ...]]",
        ),
        (
            "psycopg.crdb.CrdbConnection.connect(row_factory=rows.tuple_row)",
            "psycopg.crdb.CrdbConnection[Tuple[Any, ...]]",
        ),
        (
            "psycopg.crdb.CrdbConnection.connect(row_factory=rows.dict_row)",
            "psycopg.crdb.CrdbConnection[Dict[str, Any]]",
        ),
        (
            "await psycopg.crdb.AsyncCrdbConnection.connect()",
            "psycopg.crdb.AsyncCrdbConnection[Tuple[Any, ...]]",
        ),
        (
            "await psycopg.crdb.AsyncCrdbConnection.connect(row_factory=rows.dict_row)",
            "psycopg.crdb.AsyncCrdbConnection[Dict[str, Any]]",
        ),
    ],
)
def test_connection_type(conn, type, mypy):
    stmts = f"obj = {conn}"
    _test_reveal_crdb(stmts, type, mypy)


def _test_reveal_crdb(stmts, type, mypy):
    stmts = f"""\
import psycopg.crdb
{stmts}
"""
    _test_reveal(stmts, type, mypy)



"""
commands pipeline management
"""

# Copyright (C) 2021 The Psycopg Team

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, TypeAlias
from collections import deque

from . import errors as e
from . import pq
from .abc import PipelineCommand, PQGen
from .pq.misc import connection_summary
from .generators import fetch_many, pipeline_communicate, send
from ._capabilities import capabilities

if TYPE_CHECKING:
    from .pq.abc import PGresult
    from ._preparing import Key, Prepare  # noqa: F401
    from ._cursor_base import BaseCursor  # noqa: F401
    from ._connection_base import BaseConnection


PendingResult: TypeAlias = (
    "tuple[BaseCursor[Any, Any], tuple[Key, Prepare, bytes] | None] | None"
)

FATAL_ERROR = pq.ExecStatus.FATAL_ERROR
PIPELINE_ABORTED = pq.ExecStatus.PIPELINE_ABORTED
BAD = pq.ConnStatus.BAD

ACTIVE = pq.TransactionStatus.ACTIVE

logger = logging.getLogger("psycopg")


class BasePipeline:
    command_queue: deque[PipelineCommand]
    result_queue: deque[PendingResult]

    def __init__(self, conn: BaseConnection[Any]) -> None:
        self._conn = conn
        self.pgconn = conn.pgconn
        self.command_queue = deque[PipelineCommand]()
        self.result_queue = deque[PendingResult]()
        self.level = 0

    def __repr__(self) -> str:
        cls = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        info = connection_summary(self._conn.pgconn)
        return f"<{cls} {info} at 0x{id(self):x}>"

    @property
    def status(self) -> pq.PipelineStatus:
        return pq.PipelineStatus(self.pgconn.pipeline_status)

    @classmethod
    def is_supported(cls) -> bool:
        """Return `!True` if the psycopg libpq wrapper supports pipeline mode."""
        return capabilities.has_pipeline()

    def _enter_gen(self) -> PQGen[None]:
        capabilities.has_pipeline(check=True)
        if self.level == 0:
            self.pgconn.enter_pipeline_mode()
        elif self.command_queue or self.pgconn.transaction_status == ACTIVE:
            # Nested pipeline case.
            #  Transaction might be ACTIVE when the pipeline uses an "implicit
            #  transaction", typically in autocommit mode. But when entering a
            #  Psycopg transaction(), we expect the IDLE state. By sync()-ing,
            #  we make sure all previous commands are completed and the
            #  transaction gets back to IDLE.
            yield from self._sync_gen()
        self.level += 1

    def _exit(self, exc: BaseException | None) -> None:
        self.level -= 1
        if self.level == 0 and self.pgconn.status != BAD:
            try:
                self.pgconn.exit_pipeline_mode()
            except e.OperationalError as exc2:
                # Notice that this error might be pretty irrecoverable. It
                # happens on COPY, for instance: even if sync succeeds, exiting
                # fails with "cannot exit pipeline mode with uncollected results"
                if exc:
                    logger.warning("error ignored exiting %r: %s", self, exc2)
                else:
                    raise exc2.with_traceback(None)

    def _sync_gen(self) -> PQGen[None]:
        self._enqueue_sync()
        yield from self._communicate_gen()
        yield from self._fetch_gen(flush=False)

    def _exit_gen(self) -> PQGen[None]:
        """
        Exit current pipeline by sending a Sync and fetch back all remaining results.
        """
        try:
            self._enqueue_sync()
            yield from self._communicate_gen()
        finally:
            yield from self._fetch_gen(flush=True)

    def _communicate_gen(self) -> PQGen[None]:
        """Communicate with pipeline to send commands and possibly fetch
        results, which are then processed.
        """
        fetched = yield from pipeline_communicate(self.pgconn, self.command_queue)
        exception = None
        for results in fetched:
            queued = self.result_queue.popleft()
            try:
                self._process_results(queued, results)
            except e.Error as exc:
                if exception is None:
                    exception = exc
        if exception is not None:
            raise exception

    def _fetch_gen(self, *, flush: bool) -> PQGen[None]:
        """Fetch available results from the connection and process them with
        pipeline queued items.

        If 'flush' is True, a PQsendFlushRequest() is issued in order to make
        sure results can be fetched. Otherwise, the caller may emit a
        PQpipelineSync() call to ensure the output buffer gets flushed before
        fetching.
        """
        if not self.result_queue:
            return

        if flush:
            self.pgconn.send_flush_request()
            yield from send(self.pgconn)

        exception = None
        while self.result_queue:
            if not (results := (yield from fetch_many(self.pgconn))):
                # No more results to fetch, but there may still be pending
                # commands.
                break
            queued = self.result_queue.popleft()
            try:
                self._process_results(queued, results)
            except e.Error as exc:
                if exception is None:
                    exception = exc
        if exception is not None:
            raise exception

    def _process_results(self, queued: PendingResult, results: list[PGresult]) -> None:
        """Process a results set fetched from the current pipeline.

        This matches 'results' with its respective element in the pipeline
        queue. For commands (None value in the pipeline queue), results are
        checked directly. For prepare statement creation requests, update the
        cache. Otherwise, results are attached to their respective cursor.
        """
        if queued is None:
            (result,) = results
            if result.status == FATAL_ERROR:
                raise e.error_from_result(result, encoding=self.pgconn._encoding)
            elif result.status == PIPELINE_ABORTED:
                raise e.PipelineAborted("pipeline aborted")
        else:
            cursor, prepinfo = queued
            if prepinfo:
                key, prep, name = prepinfo
                # Update the prepare state of the query.
                cursor._conn._prepared.validate(key, prep, name, results)
            cursor._check_results(results)
            cursor._set_results(results)

    def _enqueue_sync(self) -> None:
        """Enqueue a PQpipelineSync() command."""
        self.command_queue.append(self.pgconn.pipeline_sync)
        self.result_queue.append(None)



from copy import deepcopy

import pytest

from psycopg.crdb import CrdbConnection, adapters
from psycopg.adapt import PyFormat, Transformer
from psycopg.postgres import types as builtins
from psycopg.types.array import ListDumper

from ..test_adapt import MyStr, make_bin_dumper, make_bin_loader, make_dumper
from ..test_adapt import make_loader

pytestmark = pytest.mark.crdb


@pytest.mark.parametrize("fmt_in", PyFormat)
def test_return_untyped(conn, fmt_in):
    # Analyze and check for changes using strings in untyped/typed contexts
    cur = conn.cursor()
    # Currently string are passed as text oid to CockroachDB, unlike Postgres,
    # to which strings are passed as unknown. This is because CRDB doesn't
    # allow the unknown oid to be emitted; execute("SELECT %s", ["str"]) raises
    # an error. However, unlike PostgreSQL, text can be cast to any other type.
    cur.execute(f"select %{fmt_in.value}, %{fmt_in.value}", ["hello", 10])
    assert cur.fetchone() == ("hello", 10)

    cur.execute("create table testjson(data jsonb)")
    cur.execute(f"insert into testjson (data) values (%{fmt_in.value})", ["{}"])
    assert cur.execute("select data from testjson").fetchone() == ({},)


def test_str_list_dumper_text(conn):
    t = Transformer(conn)
    dstr = t.get_dumper([""], PyFormat.TEXT)
    assert isinstance(dstr, ListDumper)
    assert dstr.oid == builtins["text"].array_oid
    assert dstr.sub_dumper and dstr.sub_dumper.oid == builtins["text"].oid


@pytest.fixture
def crdb_adapters():
    """Restore the crdb adapters after a test has changed them."""
    dumpers = deepcopy(adapters._dumpers)
    dumpers_by_oid = deepcopy(adapters._dumpers_by_oid)
    loaders = deepcopy(adapters._loaders)
    types = list(adapters.types)

    yield None

    adapters._dumpers = dumpers
    adapters._dumpers_by_oid = dumpers_by_oid
    adapters._loaders = loaders
    adapters.types.clear()
    for t in types:
        adapters.types.add(t)


def test_dump_global_ctx(dsn, crdb_adapters, pgconn):
    adapters.register_dumper(MyStr, make_bin_dumper("gb"))
    adapters.register_dumper(MyStr, make_dumper("gt"))
    with CrdbConnection.connect(dsn) as conn:
        cur = conn.execute("select %s", [MyStr("hello")])
        assert cur.fetchone() == ("hellogt",)
        cur = conn.execute("select %b", [MyStr("hello")])
        assert cur.fetchone() == ("hellogb",)
        cur = conn.execute("select %t", [MyStr("hello")])
        assert cur.fetchone() == ("hellogt",)


def test_load_global_ctx(dsn, crdb_adapters):
    adapters.register_loader("text", make_loader("gt"))
    adapters.register_loader("text", make_bin_loader("gb"))
    with CrdbConnection.connect(dsn) as conn:
        cur = conn.cursor(binary=False).execute("select 'hello'::text")
        assert cur.fetchone() == ("hellogt",)
        cur = conn.cursor(binary=True).execute("select 'hello'::text")
        assert cur.fetchone() == ("hellogb",)



from __future__ import annotations

import pytest

import psycopg
from psycopg.conninfo import conninfo_to_dict

from .test_dns import import_dnspython

pytestmark = [pytest.mark.dns]

samples_ok = [
    ("", "", None),
    ("host=_pg._tcp.foo.com", "host=db1.example.com port=5432", None),
    ("", "host=db1.example.com port=5432", {"PGHOST": "_pg._tcp.foo.com"}),
    (
        "host=foo.com,_pg._tcp.foo.com",
        "host=foo.com,db1.example.com port=,5432",
        None,
    ),
    (
        "host=_pg._tcp.dot.com,foo.com,_pg._tcp.foo.com",
        "host=foo.com,db1.example.com port=,5432",
        None,
    ),
    (
        "host=_pg._tcp.bar.com",
        "host=db1.example.com,db4.example.com,db3.example.com,db2.example.com"
        " port=5432,5432,5433,5432",
        None,
    ),
    (
        "host=service.foo.com port=srv",
        "host=service.example.com port=15432",
        None,
    ),
    # No resolution
    (
        "host=_pg._tcp.foo.com hostaddr=1.1.1.1",
        "host=_pg._tcp.foo.com hostaddr=1.1.1.1",
        None,
    ),
]


@pytest.mark.flakey("random weight order, might cause wrong order")
@pytest.mark.parametrize("conninfo, want, env", samples_ok)
def test_srv(conninfo, want, env, fake_srv, setpgenv):
    setpgenv(env)
    params = conninfo_to_dict(conninfo)
    params = psycopg._dns.resolve_srv(params)  # type: ignore[attr-defined]
    assert conninfo_to_dict(want) == params


@pytest.mark.anyio
@pytest.mark.parametrize("conninfo, want, env", samples_ok)
async def test_srv_async(conninfo, want, env, afake_srv, setpgenv):
    setpgenv(env)
    params = conninfo_to_dict(conninfo)
    params = await psycopg._dns.resolve_srv_async(params)  # type: ignore[attr-defined]
    assert conninfo_to_dict(want) == params


samples_bad = [
    ("host=_pg._tcp.dot.com", None),
    ("host=_pg._tcp.foo.com port=1,2", None),
]


@pytest.mark.parametrize("conninfo,  env", samples_bad)
def test_srv_bad(conninfo, env, fake_srv, setpgenv):
    setpgenv(env)
    params = conninfo_to_dict(conninfo)
    with pytest.raises(psycopg.OperationalError):
        psycopg._dns.resolve_srv(params)  # type: ignore[attr-defined]


@pytest.mark.anyio
@pytest.mark.parametrize("conninfo,  env", samples_bad)
async def test_srv_bad_async(conninfo, env, afake_srv, setpgenv):
    setpgenv(env)
    params = conninfo_to_dict(conninfo)
    with pytest.raises(psycopg.OperationalError):
        await psycopg._dns.resolve_srv_async(params)  # type: ignore[attr-defined]


@pytest.fixture
def fake_srv(monkeypatch):
    f = get_fake_srv_function(monkeypatch)
    monkeypatch.setattr(
        psycopg._dns.resolver,  # type: ignore[attr-defined]
        "resolve",
        f,
    )


@pytest.fixture
def afake_srv(monkeypatch):
    f = get_fake_srv_function(monkeypatch)

    async def af(qname, rdtype):
        return f(qname, rdtype)

    monkeypatch.setattr(
        psycopg._dns.async_resolver,  # type: ignore[attr-defined]
        "resolve",
        af,
    )


def get_fake_srv_function(monkeypatch):
    import_dnspython()

    from dns.exception import DNSException
    from dns.rdtypes.IN.A import A
    from dns.rdtypes.IN.SRV import SRV

    fake_hosts = {
        ("_pg._tcp.dot.com", "SRV"): ["0 0 5432 ."],
        ("_pg._tcp.foo.com", "SRV"): ["0 0 5432 db1.example.com."],
        ("_pg._tcp.bar.com", "SRV"): [
            "1 0 5432 db2.example.com.",
            "1 255 5433 db3.example.com.",
            "0 0 5432 db1.example.com.",
            "1 65535 5432 db4.example.com.",
        ],
        ("service.foo.com", "SRV"): ["0 0 15432 service.example.com."],
    }

    def fake_srv_(qname, rdtype):
        try:
            ans = fake_hosts[qname, rdtype]
        except KeyError:
            raise DNSException(f"unknown test host: {qname} {rdtype}")
        rv: list[A | SRV] = []

        if rdtype == "A":
            for entry in ans:
                rv.append(A("IN", "A", entry))
        else:
            for entry in ans:
                pri, w, port, target = entry.split()
                rv.append(SRV("IN", "SRV", int(pri), int(w), int(port), target))

        return rv

    return fake_srv_



import pytest

from ..conftest import asyncio_options


@pytest.fixture(
    params=[pytest.param(("asyncio", asyncio_options.copy()), id="asyncio")],
    scope="session",
)
def anyio_backend(request):
    backend, options = request.param
    if request.config.option.loop == "uvloop":
        options["use_uvloop"] = True
    return backend, options



"""
C optimised functions for the copy system.

"""

# Copyright (C) 2020 The Psycopg Team

from libc.stdint cimport int32_t, uint16_t, uint32_t
from libc.string cimport memcpy
from cpython.sequence cimport PySequence_Fast, PySequence_Fast_GET_ITEM
from cpython.sequence cimport PySequence_Fast_GET_SIZE
from cpython.bytearray cimport PyByteArray_AS_STRING, PyByteArray_FromStringAndSize
from cpython.bytearray cimport PyByteArray_GET_SIZE, PyByteArray_Resize
from cpython.memoryview cimport PyMemoryView_FromObject

from psycopg_c.pq cimport ViewBuffer
from psycopg_c._psycopg cimport endian

from psycopg import errors as e


cdef int32_t _binary_null = -1


cdef object _format_row_binary(object row, Transformer tx, bytearray out):
    """Convert a row of adapted data to the data to send for binary copy"""
    cdef row_fast = PySequence_Fast(row, "'row' is not a valid sequence")
    cdef Py_ssize_t rowlen = PySequence_Fast_GET_SIZE(row_fast)
    cdef uint16_t berowlen = endian.htobe16(<int16_t>rowlen)

    # offset in 'out' where to write
    cdef Py_ssize_t pos = PyByteArray_GET_SIZE(out)

    cdef char *target = CDumper.ensure_size(out, pos, sizeof(berowlen))

    # Write the number of fields as network-order 16 bits
    memcpy(target, <void *>&berowlen, sizeof(berowlen))
    pos += sizeof(berowlen)

    cdef Py_ssize_t size
    cdef uint32_t besize
    cdef char *buf
    cdef int i
    cdef PyObject *fmt = <PyObject *>PG_BINARY
    cdef PyObject *row_dumper

    if not tx._row_dumpers:
        tx._row_dumpers = PyList_New(rowlen)

    dumpers = tx._row_dumpers
    if PyList_GET_SIZE(dumpers) != rowlen:
        raise e.DataError(f"expected {len(dumpers)} values in row, got {rowlen}")

    cdef PyObject *item
    for i in range(rowlen):
        item = PySequence_Fast_GET_ITEM(row_fast, i)
        if item is <PyObject *>None:
            _append_binary_none(out, &pos)
            continue

        row_dumper = PyList_GET_ITEM(dumpers, i)
        if not row_dumper:
            row_dumper = tx.get_row_dumper(item, fmt)
            Py_INCREF(<object>row_dumper)
            PyList_SET_ITEM(dumpers, i, <object>row_dumper)

        if (<RowDumper>row_dumper).cdumper is not None:
            # A cdumper can resize if necessary and copy in place
            size = (<RowDumper>row_dumper).cdumper.cdump(
                <object>item, out, pos + sizeof(besize))
            # Also add the size of the item, before the item
            besize = endian.htobe32(<int32_t>size)
            target = PyByteArray_AS_STRING(out)  # might have been moved by cdump
            memcpy(target + pos, <void *>&besize, sizeof(besize))
        else:
            # A Python dumper, gotta call it and extract its juices
            b = PyObject_CallFunctionObjArgs(
                (<RowDumper>row_dumper).dumpfunc, item, NULL)
            if b is None:
                _append_binary_none(out, &pos)
                continue
            else:
                _buffer_as_string_and_size(b, &buf, &size)
                target = CDumper.ensure_size(out, pos, size + sizeof(besize))
                besize = endian.htobe32(<int32_t>size)
                memcpy(target, <void *>&besize, sizeof(besize))
                memcpy(target + sizeof(besize), buf, size)

        pos += size + sizeof(besize)

    # Resize to the final size
    PyByteArray_Resize(out, pos)


def format_row_binary(row: Sequence[Any], tx: Transformer, out: bytearray) -> None:
    cdef Py_ssize_t size = PyByteArray_GET_SIZE(out)

    try:
        _format_row_binary(row, tx, out)
    except Exception as e:
        # Restore the input bytearray to the size it was before entering here
        # to avoid potentially passing junk to copy.
        PyByteArray_Resize(out, size)
        raise e


cdef int _append_binary_none(bytearray out, Py_ssize_t *pos) except -1:
    cdef char *target
    target = CDumper.ensure_size(out, pos[0], sizeof(_binary_null))
    memcpy(target, <void *>&_binary_null, sizeof(_binary_null))
    pos[0] += sizeof(_binary_null)
    return 0


cdef object _format_row_text(object row, Transformer tx, bytearray out):
    # offset in 'out' where to write
    cdef Py_ssize_t pos = PyByteArray_GET_SIZE(out)
    cdef row_fast = PySequence_Fast(row, "'row' is not a valid sequence")

    # exit early, if the row is empty
    cdef Py_ssize_t rowlen = PySequence_Fast_GET_SIZE(row_fast)
    if rowlen == 0:
        PyByteArray_Resize(out, pos + 1)
        out[pos] = b"\n"
        return

    cdef Py_ssize_t size, tmpsize
    cdef char *buf
    cdef int i, j
    cdef unsigned char *target
    cdef int nesc
    cdef int with_tab
    cdef PyObject *fmt = <PyObject *>PG_TEXT
    cdef PyObject *row_dumper

    dumpers = tx._row_dumpers
    if dumpers and PyList_GET_SIZE(dumpers) != rowlen:
        raise e.DataError(f"expected {len(dumpers)} values in row, got {rowlen}")

    cdef PyObject *item
    for i in range(rowlen):
        # Include the tab before the data, so it gets included in the resizes
        with_tab = i > 0

        item = PySequence_Fast_GET_ITEM(row_fast, i)
        if item == <PyObject *>None:
            _append_text_none(out, &pos, with_tab)
            continue

        if dumpers:
            # pinned dumpers from set_types are authoritative
            row_dumper = PyList_GET_ITEM(dumpers, i)
        else:
            # no pinned dumpers, thus free value dumping
            row_dumper = tx.get_row_dumper(item, fmt)

        if (<RowDumper>row_dumper).cdumper is not None:
            # A cdumper can resize if necessary and copy in place
            size = (<RowDumper>row_dumper).cdumper.cdump(
                <object>item, out, pos + with_tab)
            target = <unsigned char *>PyByteArray_AS_STRING(out) + pos
        else:
            # A Python dumper, gotta call it and extract its juices
            b = PyObject_CallFunctionObjArgs(
                (<RowDumper>row_dumper).dumpfunc, item, NULL)
            if b is None:
                _append_text_none(out, &pos, with_tab)
                continue
            else:
                _buffer_as_string_and_size(b, &buf, &size)
                target = <unsigned char *>CDumper.ensure_size(out, pos, size + with_tab)
                memcpy(target + with_tab, buf, size)

        # Prepend a tab to the data just written
        if with_tab:
            target[0] = b"\t"
            target += 1
            pos += 1

        # Now from pos to pos + size there is a textual representation: it may
        # contain chars to escape. Scan to find how many such chars there are.
        nesc = 0
        for j in range(size):
            if copy_escape_lut[target[j]]:
                nesc += 1

        # If there is any char to escape, walk backwards pushing the chars
        # forward and interspersing backslashes.
        if nesc > 0:
            tmpsize = size + nesc
            target = <unsigned char *>CDumper.ensure_size(out, pos, tmpsize)
            for j in range(<int>size - 1, -1, -1):
                if copy_escape_lut[target[j]]:
                    target[j + nesc] = copy_escape_lut[target[j]]
                    nesc -= 1
                    target[j + nesc] = b"\\"
                    if nesc <= 0:
                        break
                else:
                    target[j + nesc] = target[j]
            pos += tmpsize
        else:
            pos += size

    # Resize to the final size, add the newline
    PyByteArray_Resize(out, pos + 1)
    out[pos] = b"\n"


def format_row_text(row: Sequence[Any], tx: Transformer, out: bytearray) -> None:
    cdef Py_ssize_t size = PyByteArray_GET_SIZE(out)

    try:
        _format_row_text(row, tx, out)
    except Exception as e:
        # Restore the input bytearray to the size it was before entering here
        # to avoid potentially passing junk to copy.
        PyByteArray_Resize(out, size)
        raise e


cdef int _append_text_none(bytearray out, Py_ssize_t *pos, int with_tab) except -1:
    cdef char *target

    if with_tab:
        target = CDumper.ensure_size(out, pos[0], 3)
        memcpy(target, b"\t\\N", 3)
        pos[0] += 3
    else:
        target = CDumper.ensure_size(out, pos[0], 2)
        memcpy(target, b"\\N", 2)
        pos[0] += 2

    return 0


def parse_row_binary(data, tx: Transformer) -> tuple[Any, ...]:
    cdef unsigned char *ptr
    cdef Py_ssize_t bufsize
    _buffer_as_string_and_size(data, <char **>&ptr, &bufsize)
    cdef unsigned char *bufend = ptr + bufsize

    cdef uint16_t benfields
    memcpy(&benfields, ptr, sizeof(benfields))
    cdef int nfields = endian.be16toh(benfields)
    ptr += sizeof(benfields)
    cdef list row = PyList_New(nfields)

    cdef int col
    cdef int32_t belength
    cdef Py_ssize_t length

    for col in range(nfields):
        memcpy(&belength, ptr, sizeof(belength))
        ptr += sizeof(belength)
        if belength == _binary_null:
            field = None
        else:
            length = endian.be32toh(belength)
            if ptr + length > bufend:
                raise e.DataError("bad copy data: length exceeding data")
            field = PyMemoryView_FromObject(
                ViewBuffer._from_buffer(data, ptr, length))
            ptr += length

        Py_INCREF(field)
        PyList_SET_ITEM(row, col, field)

    return tx.load_sequence(row)


def parse_row_text(data, tx: Transformer) -> tuple[Any, ...]:
    cdef unsigned char *fstart
    cdef Py_ssize_t size
    _buffer_as_string_and_size(data, <char **>&fstart, &size)

    # politely assume that the number of fields will be what in the result
    cdef int nfields = tx._nfields
    cdef list row = PyList_New(nfields)

    cdef unsigned char *fend
    cdef unsigned char *rowend = fstart + size
    cdef unsigned char *src
    cdef unsigned char *tgt
    cdef int col
    cdef int num_bs

    for col in range(nfields):
        fend = fstart
        num_bs = 0
        # Scan to the end of the field, remember if you see any backslash
        while fend[0] != b'\t' and fend[0] != b'\n' and fend < rowend:
            if fend[0] == b'\\':
                num_bs += 1
                # skip the next char to avoid counting escaped backslashes twice
                fend += 1
            fend += 1

        # Check if we stopped for the right reason
        if fend >= rowend:
            raise e.DataError("bad copy data: field delimiter not found")
        elif fend[0] == b'\t' and col == nfields - 1:
            raise e.DataError("bad copy data: got a tab at the end of the row")
        elif fend[0] == b'\n' and col != nfields - 1:
            raise e.DataError(
                "bad copy format: got a newline before the end of the row")

        # Is this a NULL?
        if fend - fstart == 2 and fstart[0] == b'\\' and fstart[1] == b'N':
            field = None

        # Is this a field with no backslash?
        elif num_bs == 0:
            # Nothing to unescape: we don't need a copy
            field = PyMemoryView_FromObject(
                ViewBuffer._from_buffer(data, fstart, fend - fstart))

        # This is a field containing backslashes
        else:
            # We need a copy of the buffer to unescape
            field = PyByteArray_FromStringAndSize("", 0)
            PyByteArray_Resize(field, fend - fstart - num_bs)
            tgt = <unsigned char *>PyByteArray_AS_STRING(field)
            src = fstart
            while (src < fend):
                if src[0] != b'\\':
                    tgt[0] = src[0]
                else:
                    src += 1
                    tgt[0] = copy_unescape_lut[src[0]]
                src += 1
                tgt += 1

        Py_INCREF(field)
        PyList_SET_ITEM(row, col, field)

        # Start of the field
        fstart = fend + 1

    # Convert the array of buffers into Python objects
    return tx.load_sequence(row)


cdef extern from *:
    """
/* handle chars to (un)escape in text copy representation */
/* '\b', '\t', '\n', '\v', '\f', '\r', '\\' */

/* Escaping chars */
static const char copy_escape_lut[] = {
  0,   0,   0,   0,   0,   0,   0,   0,  98, 116, 110, 118, 102, 114,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  92,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
};

/* Conversion of escaped to unescaped chars */
static const char copy_unescape_lut[] = {
  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,  15,
 16,  17,  18,  19,  20,  21,  22,  23,  24,  25,  26,  27,  28,  29,  30,  31,
 32,  33,  34,  35,  36,  37,  38,  39,  40,  41,  42,  43,  44,  45,  46,  47,
 48,  49,  50,  51,  52,  53,  54,  55,  56,  57,  58,  59,  60,  61,  62,  63,
 64,  65,  66,  67,  68,  69,  70,  71,  72,  73,  74,  75,  76,  77,  78,  79,
 80,  81,  82,  83,  84,  85,  86,  87,  88,  89,  90,  91,  92,  93,  94,  95,
 96,  97,   8,  99, 100, 101,  12, 103, 104, 105, 106, 107, 108, 109,  10, 111,
112, 113,  13, 115,   9, 117,  11, 119, 120, 121, 122, 123, 124, 125, 126, 127,
128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143,
144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159,
160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175,
176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191,
192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207,
208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223,
224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239,
240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255,
};
    """
    const char[256] copy_escape_lut
    const char[256] copy_unescape_lut



"""
Helper object to transform values between Python and PostgreSQL

Cython implementation: can access to lower level C features without creating
too many temporary Python objects and performing less memory copying.

"""

# Copyright (C) 2020 The Psycopg Team

cimport cython
from cpython.ref cimport Py_INCREF
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.list cimport PyList_GET_ITEM, PyList_GET_SIZE, PyList_New, PyList_SET_ITEM
from cpython.bytes cimport PyBytes_AS_STRING
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM
from cpython.object cimport PyObject, PyObject_CallFunctionObjArgs
from cpython.sequence cimport PySequence_Fast, PySequence_Fast_GET_ITEM
from cpython.sequence cimport PySequence_Fast_GET_SIZE

from typing import Sequence

from psycopg import errors as e
from psycopg.pq import Format as PqFormat
from psycopg.rows import Row
from psycopg._encodings import conn_encoding

NoneType = type(None)

# internal structure: you are not supposed to know this. But it's worth some
# 10% of the innermost loop, so I'm willing to ask for forgiveness later...

ctypedef struct PGresAttValue:
    int len
    char *value

ctypedef struct pg_result_int:
    # NOTE: it would be advised that we don't know this structure's content
    int ntups
    int numAttributes
    libpq.PGresAttDesc *attDescs
    PGresAttValue **tuples
    # ...more members, which we ignore


@cython.freelist(16)
cdef class RowLoader:
    cdef CLoader cloader
    cdef object pyloader
    cdef object loadfunc


@cython.freelist(16)
cdef class RowDumper:
    cdef CDumper cdumper
    cdef object pydumper
    cdef object dumpfunc
    cdef object oid
    cdef object format


cdef class Transformer:
    """
    An object that can adapt efficiently between Python and PostgreSQL.

    The life cycle of the object is the query, so it is assumed that attributes
    such as the server version or the connection encoding will not change. The
    object have its state so adapting several values of the same type can be
    optimised.

    """

    cdef readonly object connection
    cdef readonly object adapters
    cdef readonly object types
    cdef readonly object formats
    cdef str _encoding
    cdef int _none_oid

    # mapping class -> Dumper instance (auto, text, binary)
    cdef dict _auto_dumpers
    cdef dict _text_dumpers
    cdef dict _binary_dumpers

    # mapping oid -> Loader instance (text, binary)
    cdef dict _text_loaders
    cdef dict _binary_loaders

    # mapping oid -> Dumper instance (text, binary)
    cdef dict _oid_text_dumpers
    cdef dict _oid_binary_dumpers

    cdef pq.PGresult _pgresult
    cdef int _nfields, _ntuples
    cdef list _row_dumpers
    cdef list _row_loaders

    cdef dict _oid_types

    def __cinit__(self, context: "AdaptContext" | None = None):
        if context is not None:
            self.adapters = context.adapters
            self.connection = context.connection
        else:
            from psycopg import postgres
            self.adapters = postgres.adapters
            self.connection = None

        self.types = self.formats = None
        self._none_oid = -1

    @classmethod
    def from_context(cls, context: "AdaptContext" | None):
        """
        Return a Transformer from an AdaptContext.

        If the context is a Transformer instance, just return it.
        """
        return _tx_from_context(context)

    @property
    def encoding(self) -> str:
        if not self._encoding:
            self._encoding = conn_encoding(self.connection)
        return self._encoding

    @property
    def pgresult(self) -> PGresult | None:
        return self._pgresult

    cpdef set_pgresult(
        self,
        pq.PGresult result,
        object set_loaders = True,
        object format = None
    ):
        self._pgresult = result

        if result is None:
            self._nfields = self._ntuples = 0
            if set_loaders:
                self._row_loaders = []
            return

        cdef libpq.PGresult *res = self._pgresult._pgresult_ptr
        self._nfields = libpq.PQnfields(res)
        self._ntuples = libpq.PQntuples(res)

        if not set_loaders:
            return

        if not self._nfields:
            self._row_loaders = []
            return

        if format is None:
            format = libpq.PQfformat(res, 0)

        cdef list loaders = PyList_New(self._nfields)
        cdef PyObject *row_loader
        cdef object oid

        cdef int i
        for i in range(self._nfields):
            oid = libpq.PQftype(res, i)
            row_loader = self._c_get_loader(<PyObject *>oid, <PyObject *>format)
            Py_INCREF(<object>row_loader)
            PyList_SET_ITEM(loaders, i, <object>row_loader)

        self._row_loaders = loaders

    def set_dumper_types(self, types: Sequence[int], format: PqFormat) -> None:
        # NOTE: impl detail - final _row_dumpers must be a list type
        # (assumed by format_row_binary and format_row_text)
        cdef Py_ssize_t ntypes = len(types)
        dumpers = PyList_New(ntypes)
        cdef int i
        for i in range(ntypes):
            oid = types[i]
            dumper_ptr = self.get_dumper_by_oid(
                <PyObject *>oid, <PyObject *>format)
            Py_INCREF(<object>dumper_ptr)
            PyList_SET_ITEM(dumpers, i, <object>dumper_ptr)

        self._row_dumpers = dumpers
        self.types = tuple(types)
        self.formats = [format] * ntypes

    def set_loader_types(self, types: Sequence[int], format: PqFormat) -> None:
        cdef types_fast = PySequence_Fast(types, "'types' is not a valid sequence")
        cdef Py_ssize_t ntypes = PySequence_Fast_GET_SIZE(types_fast)
        cdef list loaders = PyList_New(ntypes)

        # these are used more as Python object than C
        cdef PyObject *oid
        cdef PyObject *row_loader
        for i in range(ntypes):
            oid = PySequence_Fast_GET_ITEM(types_fast, i)
            row_loader = self._c_get_loader(oid, <PyObject *>format)
            Py_INCREF(<object>row_loader)
            PyList_SET_ITEM(loaders, i, <object>row_loader)

        self._row_loaders = loaders

    cpdef as_literal(self, obj):
        cdef PyObject *row_dumper = self.get_row_dumper(
            <PyObject *>obj, <PyObject *>PG_TEXT)

        if (<RowDumper>row_dumper).cdumper is not None:
            dumper = (<RowDumper>row_dumper).cdumper
        else:
            dumper = (<RowDumper>row_dumper).pydumper

        rv = dumper.quote(obj)
        oid = dumper.oid
        # If the result is quoted and the oid not unknown or text,
        # add an explicit type cast.
        # Check the last char because the first one might be 'E'.
        if oid and oid != oids.TEXT_OID and rv and rv[-1] == 39:
            if self._oid_types is None:
                self._oid_types = {}
            type_ptr = PyDict_GetItem(<object>self._oid_types, oid)
            if type_ptr == NULL:
                type_sql = b""
                ti = self.adapters.types.get(oid)
                if ti is not None:
                    if oid < 8192:
                        # builtin: prefer "timestamptz" to "timestamp with time zone"
                        type_sql = ti.name.encode(self.encoding)
                    else:
                        type_sql = ti.regtype.encode(self.encoding)
                    if oid == ti.array_oid:
                        type_sql += b"[]"

                type_ptr = <PyObject *>type_sql
                PyDict_SetItem(<object>self._oid_types, oid, type_sql)

            if <object>type_ptr:
                rv = b"%s::%s" % (rv, <object>type_ptr)

        if not isinstance(rv, bytes):
            rv = bytes(rv)

        return rv

    def get_dumper(self, obj, format) -> "Dumper":
        cdef PyObject *row_dumper = self.get_row_dumper(
            <PyObject *>obj, <PyObject *>format)
        return (<RowDumper>row_dumper).pydumper

    cdef PyObject *get_row_dumper(self, PyObject *obj, PyObject *fmt) except NULL:
        """
        Return a borrowed reference to the RowDumper for the given obj/fmt.
        """
        # Fast path: return a Dumper class already instantiated from the same type
        cdef PyObject *cache
        cdef PyObject *ptr
        cdef PyObject *ptr1
        cdef RowDumper row_dumper

        # Normally, the type of the object dictates how to dump it
        key = type(<object>obj)

        # Establish where would the dumper be cached
        bfmt = PyUnicode_AsUTF8String(<object>fmt)
        cdef char cfmt = PyBytes_AS_STRING(bfmt)[0]
        if cfmt == b's':
            if self._auto_dumpers is None:
                self._auto_dumpers = {}
            cache = <PyObject *>self._auto_dumpers
        elif cfmt == b'b':
            if self._binary_dumpers is None:
                self._binary_dumpers = {}
            cache = <PyObject *>self._binary_dumpers
        elif cfmt == b't':
            if self._text_dumpers is None:
                self._text_dumpers = {}
            cache = <PyObject *>self._text_dumpers
        else:
            raise ValueError(
                f"format should be a psycopg.adapt.Format, not {<object>fmt}")

        # Reuse an existing Dumper class for objects of the same type
        ptr = PyDict_GetItem(<object>cache, key)
        if ptr == NULL:
            dcls = PyObject_CallFunctionObjArgs(
                self.adapters.get_dumper, <PyObject *>key, fmt, NULL)
            dumper = PyObject_CallFunctionObjArgs(
                dcls, <PyObject *>key, <PyObject *>self, NULL)

            row_dumper = _as_row_dumper(dumper)
            PyDict_SetItem(<object>cache, key, row_dumper)
            ptr = <PyObject *>row_dumper

        # Check if the dumper requires an upgrade to handle this specific value
        if (<RowDumper>ptr).cdumper is not None:
            key1 = (<RowDumper>ptr).cdumper.get_key(<object>obj, <object>fmt)
        else:
            key1 = PyObject_CallFunctionObjArgs(
                (<RowDumper>ptr).pydumper.get_key, obj, fmt, NULL)
        if key1 is key:
            return ptr

        # If it does, ask the dumper to create its own upgraded version
        ptr1 = PyDict_GetItem(<object>cache, key1)
        if ptr1 != NULL:
            return ptr1

        if (<RowDumper>ptr).cdumper is not None:
            dumper = (<RowDumper>ptr).cdumper.upgrade(<object>obj, <object>fmt)
        else:
            dumper = PyObject_CallFunctionObjArgs(
                (<RowDumper>ptr).pydumper.upgrade, obj, fmt, NULL)

        row_dumper = _as_row_dumper(dumper)
        PyDict_SetItem(<object>cache, key1, row_dumper)
        return <PyObject *>row_dumper

    cdef PyObject *get_dumper_by_oid(self, PyObject *oid, PyObject *fmt) except NULL:
        """
        Return a borrowed reference to the RowDumper for the given oid/fmt.
        """
        cdef PyObject *ptr
        cdef PyObject *cache
        cdef RowDumper row_dumper

        # Establish where would the dumper be cached
        cdef int cfmt = <object>fmt
        if cfmt == 0:
            if self._oid_text_dumpers is None:
                self._oid_text_dumpers = {}
            cache = <PyObject *>self._oid_text_dumpers
        elif cfmt == 1:
            if self._oid_binary_dumpers is None:
                self._oid_binary_dumpers = {}
            cache = <PyObject *>self._oid_binary_dumpers
        else:
            raise ValueError(
                f"format should be a psycopg.pq.Format, not {<object>fmt}")

        # Reuse an existing Dumper class for objects of the same type
        ptr = PyDict_GetItem(<object>cache, <object>oid)
        if ptr == NULL:
            dcls = PyObject_CallFunctionObjArgs(
                self.adapters.get_dumper_by_oid, oid, fmt, NULL)
            dumper = PyObject_CallFunctionObjArgs(
                dcls, <PyObject *>NoneType, <PyObject *>self, NULL)

            row_dumper = _as_row_dumper(dumper)
            PyDict_SetItem(<object>cache, <object>oid, row_dumper)
            ptr = <PyObject *>row_dumper

        return ptr

    cpdef dump_sequence(self, object params, object formats):
        cdef int i
        cdef PyObject *dumper_ptr  # borrowed pointer to row dumper
        cdef object dumped
        cdef Py_ssize_t size

        cdef params_fast = PySequence_Fast(
            params, "'params' is not a valid sequence")
        cdef formats_fast = PySequence_Fast(
            formats, "'formats' is not a valid sequence")

        cdef Py_ssize_t nparams = PySequence_Fast_GET_SIZE(params_fast)
        cdef list out = PyList_New(nparams)
        cdef PyObject *param

        dumpers = self._row_dumpers
        if dumpers:
            for i in range(nparams):
                param = PySequence_Fast_GET_ITEM(params_fast, i)
                if param != <PyObject *>None:
                    dumper_ptr = PyList_GET_ITEM(dumpers, i)
                    if (<RowDumper>dumper_ptr).cdumper is not None:
                        dumped = PyByteArray_FromStringAndSize("", 0)
                        size = (<RowDumper>dumper_ptr).cdumper.cdump(
                            <object>param, <bytearray>dumped, 0)
                        PyByteArray_Resize(dumped, size)
                    else:
                        dumped = PyObject_CallFunctionObjArgs(
                            (<RowDumper>dumper_ptr).dumpfunc, param, NULL)
                else:
                    dumped = None

                Py_INCREF(dumped)
                PyList_SET_ITEM(out, i, dumped)

            return out

        cdef tuple types = PyTuple_New(nparams)
        cdef list pqformats = PyList_New(nparams)
        cdef PyObject *format

        for i in range(nparams):
            param = PySequence_Fast_GET_ITEM(params_fast, i)
            if param != <PyObject *>None:
                format = PySequence_Fast_GET_ITEM(formats_fast, i)
                dumper_ptr = self.get_row_dumper(param, format)
                if (<RowDumper>dumper_ptr).cdumper is not None:
                    dumped = PyByteArray_FromStringAndSize("", 0)
                    size = (<RowDumper>dumper_ptr).cdumper.cdump(
                        <object>param, <bytearray>dumped, 0)
                    PyByteArray_Resize(dumped, size)
                else:
                    dumped = PyObject_CallFunctionObjArgs(
                        (<RowDumper>dumper_ptr).dumpfunc, param, NULL)
                oid = (<RowDumper>dumper_ptr).oid
                fmt = (<RowDumper>dumper_ptr).format
            else:
                dumped = None
                if self._none_oid < 0:
                    self._none_oid = self.adapters.get_dumper(NoneType, "s").oid
                oid = self._none_oid
                fmt = PQ_TEXT

            Py_INCREF(dumped)
            PyList_SET_ITEM(out, i, dumped)

            Py_INCREF(oid)
            PyTuple_SET_ITEM(types, i, oid)

            Py_INCREF(fmt)
            PyList_SET_ITEM(pqformats, i, fmt)

        self.types = types
        self.formats = pqformats
        return out

    def load_rows(self, int row0, int row1, object make_row) -> list[Row]:
        if self._pgresult is None:
            raise e.InterfaceError("result not set")

        if not (0 <= row0 <= self._ntuples and 0 <= row1 <= self._ntuples):
            raise e.InterfaceError(
                f"rows must be included between 0 and {self._ntuples}"
            )

        cdef libpq.PGresult *res = self._pgresult._pgresult_ptr
        # cheeky access to the internal PGresult structure
        cdef pg_result_int *ires = <pg_result_int*>res

        cdef int row
        cdef int col
        cdef PGresAttValue *attval
        cdef object record  # not 'tuple' as it would check on assignment

        cdef object records = PyList_New(row1 - row0)

        cdef PyObject *loader  # borrowed RowLoader

        row_loaders = self._row_loaders  # avoid an incref/decref per item

        for row in range(row0, row1):
            record = PyTuple_New(self._nfields)

            for col in range(self._nfields):
                attval = &(ires.tuples[row][col])
                if attval.len == -1:  # NULL_LEN
                    pyval = None
                else:
                    loader = PyList_GET_ITEM(row_loaders, col)
                    if (<RowLoader>loader).cloader is not None:
                        pyval = (<RowLoader>loader).cloader.cload(
                            attval.value, attval.len)
                    else:
                        b = PyMemoryView_FromObject(
                            ViewBuffer._from_buffer(
                                self._pgresult,
                                <unsigned char *>attval.value, attval.len))
                        pyval = PyObject_CallFunctionObjArgs(
                            (<RowLoader>loader).loadfunc, <PyObject *>b, NULL)

                Py_INCREF(pyval)
                PyTuple_SET_ITEM(record, col, pyval)

            if make_row is not tuple:
                record = PyObject_CallFunctionObjArgs(
                    make_row, <PyObject *>record, NULL)

            Py_INCREF(record)
            PyList_SET_ITEM(records, row - row0, record)

        return records

    def load_row(self, int row, object make_row) -> Row:
        if self._pgresult is None:
            raise e.InterfaceError("result not set")

        if not 0 <= row < self._ntuples:
            raise e.InterfaceError(
                f"row must be included between 0 and {self._ntuples}"
            )

        cdef libpq.PGresult *res = self._pgresult._pgresult_ptr
        # cheeky access to the internal PGresult structure
        cdef pg_result_int *ires = <pg_result_int*>res

        cdef PyObject *loader  # borrowed RowLoader
        cdef int col
        cdef PGresAttValue *attval
        cdef object record  # not 'tuple' as it would check on assignment

        record = PyTuple_New(self._nfields)
        row_loaders = self._row_loaders  # avoid an incref/decref per item

        for col in range(self._nfields):
            attval = &(ires.tuples[row][col])
            if attval.len == -1:  # NULL_LEN
                pyval = None
            else:
                loader = PyList_GET_ITEM(row_loaders, col)
                if (<RowLoader>loader).cloader is not None:
                    pyval = (<RowLoader>loader).cloader.cload(
                        attval.value, attval.len)
                else:
                    b = PyMemoryView_FromObject(
                        ViewBuffer._from_buffer(
                            self._pgresult,
                            <unsigned char *>attval.value, attval.len))
                    pyval = PyObject_CallFunctionObjArgs(
                        (<RowLoader>loader).loadfunc, <PyObject *>b, NULL)

            Py_INCREF(pyval)
            PyTuple_SET_ITEM(record, col, pyval)

        if make_row is not tuple:
            record = PyObject_CallFunctionObjArgs(
                make_row, <PyObject *>record, NULL)
        return record

    cpdef object load_sequence(self, record: Sequence[Buffer | None]):
        cdef record_fast = PySequence_Fast(record, "'record' is not a valid sequence")
        cdef Py_ssize_t nfields = PySequence_Fast_GET_SIZE(record_fast)
        cdef PyObject *loader  # borrowed RowLoader
        cdef int col
        cdef char *ptr
        cdef Py_ssize_t size
        cdef PyObject *item

        out = PyTuple_New(nfields)
        row_loaders = self._row_loaders  # avoid an incref/decref per item
        if PyList_GET_SIZE(row_loaders) != nfields:
            raise e.ProgrammingError(
                f"cannot load sequence of {nfields} items:"
                f" {len(self._row_loaders)} loaders registered")

        for col in range(nfields):
            item = PySequence_Fast_GET_ITEM(record_fast, col)
            if item == <PyObject *>None:
                pyval = None
            else:
                loader = PyList_GET_ITEM(row_loaders, col)
                if (<RowLoader>loader).cloader is not None:
                    _buffer_as_string_and_size(<object>item, &ptr, &size)
                    pyval = (<RowLoader>loader).cloader.cload(ptr, size)
                else:
                    pyval = PyObject_CallFunctionObjArgs(
                        (<RowLoader>loader).loadfunc, item, NULL)

            Py_INCREF(pyval)
            PyTuple_SET_ITEM(out, col, pyval)

        return out

    def get_loader(self, oid: int, format: PqFormat) -> "Loader":
        cdef PyObject *row_loader = self._c_get_loader(
            <PyObject *>oid, <PyObject *>format)
        return (<RowLoader>row_loader).pyloader

    cdef PyObject *_c_get_loader(self, PyObject *oid, PyObject *fmt) except NULL:
        """
        Return a borrowed reference to the RowLoader instance for given oid/fmt
        """
        cdef PyObject *ptr
        cdef PyObject *cache

        if <object>fmt == PQ_TEXT:
            if self._text_loaders is None:
                self._text_loaders = {}
            cache = <PyObject *>self._text_loaders
        elif <object>fmt == PQ_BINARY:
            if self._binary_loaders is None:
                self._binary_loaders = {}
            cache = <PyObject *>self._binary_loaders
        else:
            raise ValueError(
                f"format should be a psycopg.pq.Format, not {format}")

        ptr = PyDict_GetItem(<object>cache, <object>oid)
        if ptr != NULL:
            return ptr

        loader_cls = self.adapters.get_loader(<object>oid, <object>fmt)
        if loader_cls is None:
            loader_cls = self.adapters.get_loader(oids.INVALID_OID, <object>fmt)
            if loader_cls is None:
                raise e.InterfaceError("unknown oid loader not found")

        loader = PyObject_CallFunctionObjArgs(
            loader_cls, oid, <PyObject *>self, NULL)

        cdef RowLoader row_loader = RowLoader()
        row_loader.pyloader = loader
        row_loader.loadfunc = loader.load
        if isinstance(loader, CLoader):
            row_loader.cloader = <CLoader>loader

        PyDict_SetItem(<object>cache, <object>oid, row_loader)
        return <PyObject *>row_loader


cdef object _as_row_dumper(object dumper):
    cdef RowDumper row_dumper = RowDumper()

    row_dumper.pydumper = dumper
    row_dumper.dumpfunc = dumper.dump
    row_dumper.oid = dumper.oid
    row_dumper.format = dumper.format

    if isinstance(dumper, CDumper):
        row_dumper.cdumper = <CDumper>dumper

    return row_dumper


cdef Transformer _tx_from_context(object context):
    if isinstance(context, Transformer):
        return context
    else:
        return Transformer(context)



# WARNING: this file is auto-generated by 'async_to_sync.py'
# from the original file 'test_cursor_common_async.py'
# DO NOT CHANGE! Change the original file instead.
"""
Tests common to psycopg.Cursor and its subclasses.
"""

import weakref
import datetime as dt
from typing import Any
from contextlib import closing

import pytest
from packaging.version import parse as ver

import psycopg
from psycopg import pq, rows, sql
from psycopg.adapt import PyFormat
from psycopg.types import TypeInfo

from . import _test_cursor
from .utils import raiseif
from .acompat import gather, spawn
from .fix_crdb import crdb_encoding
from .test_adapt import make_loader
from ._test_cursor import my_row_factory, ph

execmany = _test_cursor.execmany  # avoid F811 underneath
_execmany = _test_cursor._execmany  # needed by the execmany fixture

cursor_classes = [psycopg.Cursor, psycopg.ClientCursor]
# Allow to import (not necessarily to run) the module with psycopg 3.1.
# Needed to test psycopg_pool 3.2 tests with psycopg 3.1 imported, i.e. to run
# `pytest -m pool`. (which might happen when releasing pool packages).
if ver(psycopg.__version__) >= ver("3.2.0.dev0"):
    cursor_classes.append(psycopg.RawCursor)


@pytest.fixture(params=cursor_classes)
def conn(conn, request, anyio_backend):
    conn.cursor_factory = request.param
    return conn


def test_init(conn):
    cur = conn.cursor_factory(conn)
    cur.execute("select 1")
    assert cur.fetchone() == (1,)

    conn.row_factory = rows.dict_row
    cur = conn.cursor_factory(conn)
    cur.execute("select 1 as a")
    assert cur.fetchone() == {"a": 1}


def test_init_factory(conn):
    cur = conn.cursor_factory(conn, row_factory=rows.dict_row)
    cur.execute("select 1 as a")
    assert cur.fetchone() == {"a": 1}


def test_close(conn):
    cur = conn.cursor()
    assert not cur.closed
    cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        cur.execute("select 'foo'")

    cur.close()
    assert cur.closed


def test_cursor_close_fetchone(conn):
    cur = conn.cursor()
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    cur.execute(query)
    for _ in range(5):
        cur.fetchone()

    cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        cur.fetchone()


def test_cursor_close_fetchmany(conn):
    cur = conn.cursor()
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    cur.execute(query)
    assert len(cur.fetchmany(2)) == 2

    cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        cur.fetchmany(2)


def test_cursor_close_fetchall(conn):
    cur = conn.cursor()
    assert not cur.closed

    query = "select * from generate_series(1, 10)"
    cur.execute(query)
    assert len(cur.fetchall()) == 10

    cur.close()
    assert cur.closed

    with pytest.raises(psycopg.InterfaceError):
        cur.fetchall()


def test_context(conn):
    with conn.cursor() as cur:
        assert not cur.closed

    assert cur.closed


@pytest.mark.slow
def test_weakref(conn, gc_collect):
    cur = conn.cursor()
    w = weakref.ref(cur)
    cur.close()
    del cur
    gc_collect()
    assert w() is None


def test_pgresult(conn):
    cur = conn.cursor()
    cur.execute("select 1")
    assert cur.pgresult
    cur.close()
    assert not cur.pgresult


def test_statusmessage(conn):
    cur = conn.cursor()
    assert cur.statusmessage is None

    cur.execute("select generate_series(1, 10)")
    assert cur.statusmessage == "SELECT 10"

    cur.execute("")
    assert cur.statusmessage is None

    cur.execute("create table statusmessage ()")
    assert cur.statusmessage == "CREATE TABLE"

    with pytest.raises(psycopg.ProgrammingError):
        cur.execute("wat")
    assert cur.statusmessage is None


def test_execute_sql(conn):
    cur = conn.cursor()
    cur.execute(sql.SQL("select {value}").format(value="hello"))
    assert cur.fetchone() == ("hello",)


def test_next(conn):
    cur = conn.cursor()
    cur.execute("select 1")
    assert next(cur) == (1,)
    with pytest.raises(StopIteration):
        next(cur)


def test_query_parse_cache_size(conn):
    cur = conn.cursor()
    cls = type(cur)

    # Warning: testing internal structures. Test might need refactoring with the code.
    cache: Any
    if cls is psycopg.Cursor:
        cache = psycopg._queries._query2pg
    elif cls is psycopg.ClientCursor:
        cache = psycopg._queries._query2pg_client
    elif cls is psycopg.RawCursor:
        pytest.skip("RawCursor has no query parse cache")
    else:
        assert False, cls

    cache.cache_clear()
    ci = cache.cache_info()
    h0, m0 = (ci.hits, ci.misses)
    tests = [
        (f"select 1 -- {'x' * 3500}", (), h0, m0 + 1),
        (f"select 1 -- {'x' * 3500}", (), h0 + 1, m0 + 1),
        (f"select 1 -- {'x' * 4500}", (), h0 + 1, m0 + 1),
        (f"select 1 -- {'x' * 4500}", (), h0 + 1, m0 + 1),
        (f"select 1 -- {'%s' * 40}", ("x",) * 40, h0 + 1, m0 + 2),
        (f"select 1 -- {'%s' * 40}", ("x",) * 40, h0 + 2, m0 + 2),
        (f"select 1 -- {'%s' * 60}", ("x",) * 60, h0 + 2, m0 + 2),
        (f"select 1 -- {'%s' * 60}", ("x",) * 60, h0 + 2, m0 + 2),
    ]
    for i, (query, params, hits, misses) in enumerate(tests):
        pq = cur._query_cls(psycopg.adapt.Transformer())
        pq.convert(query, params)
        ci = cache.cache_info()
        assert ci.hits == hits, f"at {i}"
        assert ci.misses == misses, f"at {i}"


def test_execute_many_results(conn):
    cur = conn.cursor()
    assert cur.nextset() is None

    rv = cur.execute("select 'foo'; select generate_series(1,3)")
    assert rv is cur
    assert cur.fetchall() == [("foo",)]
    assert cur.rowcount == 1
    assert cur.nextset()
    assert cur.fetchall() == [(1,), (2,), (3,)]
    assert cur.rowcount == 3
    assert cur.nextset() is None

    cur.close()
    assert cur.nextset() is None


def test_set_results(conn):
    cur = conn.cursor()

    with pytest.raises(IndexError):
        cur.set_result(0)

    cur.execute("select 'foo'; select generate_series(1,3)")
    assert cur.set_result(0) is cur
    assert cur.fetchall() == [("foo",)]
    assert cur.rowcount == 1

    assert cur.set_result(-1) is cur
    assert cur.fetchall() == [(1,), (2,), (3,)]
    assert cur.rowcount == 3

    with pytest.raises(IndexError):
        cur.set_result(2)

    with pytest.raises(IndexError):
        cur.set_result(-3)


def test_execute_sequence(conn):
    cur = conn.cursor()
    rv = cur.execute(ph(cur, "select %s::int, %s::text, %s::text"), [1, "foo", None])
    assert rv is cur
    assert len(cur._results) == 1
    assert cur.pgresult.get_value(0, 0) == b"1"
    assert cur.pgresult.get_value(0, 1) == b"foo"
    assert cur.pgresult.get_value(0, 2) is None
    assert cur.nextset() is None


@pytest.mark.parametrize("query", ["", " ", ";"])
def test_execute_empty_query(conn, query):
    cur = conn.cursor()
    cur.execute(query)
    assert cur.pgresult.status == pq.ExecStatus.EMPTY_QUERY
    with pytest.raises(psycopg.ProgrammingError):
        cur.fetchone()


def test_execute_type_change(conn):
    # issue #112
    conn.execute("create table bug_112 (num integer)")
    cur = conn.cursor()
    sql = ph(cur, "insert into bug_112 (num) values (%s)")
    cur.execute(sql, (1,))
    cur.execute(sql, (100000,))
    cur.execute("select num from bug_112 order by num")
    assert cur.fetchall() == [(1,), (100000,)]


def test_executemany_type_change(conn):
    conn.execute("create table bug_112 (num integer)")
    cur = conn.cursor()
    sql = ph(cur, "insert into bug_112 (num) values (%s)")
    cur.executemany(sql, [(1,), (100000,)])
    cur.execute("select num from bug_112 order by num")
    assert cur.fetchall() == [(1,), (100000,)]


@pytest.mark.parametrize(
    "query", ["copy testcopy from stdin", "copy testcopy to stdout"]
)
def test_execute_copy(conn, query):
    cur = conn.cursor()
    cur.execute("create table testcopy (id int)")
    with pytest.raises(psycopg.ProgrammingError):
        cur.execute(query)


def test_fetchone(conn):
    cur = conn.cursor()
    cur.execute(ph(cur, "select %s::int, %s::text, %s::text"), [1, "foo", None])
    assert cur.pgresult.fformat(0) == 0

    row = cur.fetchone()
    assert row == (1, "foo", None)
    row = cur.fetchone()
    assert row is None


def test_binary_cursor_execute(conn):
    with raiseif(
        conn.cursor_factory is psycopg.ClientCursor, psycopg.NotSupportedError
    ) as ex:
        cur = conn.cursor(binary=True)
        cur.execute(ph(cur, "select %s, %s"), [1, None])
    if ex:
        return

    assert cur.fetchone() == (1, None)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x01"


def test_execute_binary(conn):
    cur = conn.cursor()
    with raiseif(
        conn.cursor_factory is psycopg.ClientCursor, psycopg.NotSupportedError
    ) as ex:
        cur.execute(ph(cur, "select %s, %s"), [1, None], binary=True)
    if ex:
        return

    assert cur.fetchone() == (1, None)
    assert cur.pgresult.fformat(0) == 1
    assert cur.pgresult.get_value(0, 0) == b"\x00\x01"


def test_binary_cursor_text_override(conn):
    cur = conn.cursor(binary=True)
    cur.execute(ph(cur, "select %s, %s"), [1, None], binary=False)
    assert cur.fetchone() == (1, None)
    assert cur.pgresult.fformat(0) == 0
    assert cur.pgresult.get_value(0, 0) == b"1"


@pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
def test_query_encode(conn, encoding):
    conn.execute(f"set client_encoding to {encoding}")
    cur = conn.cursor()
    cur.execute("select '€'")
    (res,) = cur.fetchone()
    assert res == "€"


@pytest.mark.parametrize("encoding", [crdb_encoding("latin1")])
def test_query_badenc(conn, encoding):
    conn.execute(f"set client_encoding to {encoding}")
    cur = conn.cursor()
    with pytest.raises(UnicodeEncodeError):
        cur.execute("select '€'")


def test_executemany(conn, execmany):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
    )
    cur.execute("select num, data from execmany order by 1")
    rv = cur.fetchall()
    assert rv == [(10, "hello"), (20, "world")]


def test_executemany_name(conn, execmany):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%(num)s, %(data)s)"),
        [{"num": 11, "data": "hello", "x": 1}, {"num": 21, "data": "world"}],
    )
    cur.execute("select num, data from execmany order by 1")
    rv = cur.fetchall()
    assert rv == [(11, "hello"), (21, "world")]


def test_executemany_no_data(conn, execmany):
    cur = conn.cursor()
    cur.executemany(ph(cur, "insert into execmany(num, data) values (%s, %s)"), [])
    assert cur.rowcount == 0


def test_executemany_rowcount(conn, execmany):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
    )
    assert cur.rowcount == 2


@pytest.mark.parametrize("returning", [False, True])
def test_executemany_statusmessage(conn, execmany, returning):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
        returning=returning,
    )
    assert cur.rowcount == (1 if returning else 2)
    assert cur.statusmessage is not None
    assert cur.statusmessage.startswith("INSERT")

    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [],
        returning=returning,
    )
    assert cur.statusmessage is None


def test_executemany_returning(conn, execmany):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s) returning num"),
        [(10, "hello"), (20, "world")],
        returning=True,
    )
    assert cur.rowcount == 1
    assert cur.fetchone() == (10,)
    assert cur.nextset()
    assert cur.rowcount == 1
    assert cur.fetchone() == (20,)
    assert cur.nextset() is None


def test_executemany_returning_discard(conn, execmany):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s) returning num"),
        [(10, "hello"), (20, "world")],
    )
    assert cur.rowcount == 2
    with pytest.raises(psycopg.ProgrammingError):
        cur.fetchone()
    assert cur.nextset() is None


def test_executemany_no_result(conn, execmany):
    cur = conn.cursor()
    cur.executemany(
        ph(cur, "insert into execmany(num, data) values (%s, %s)"),
        [(10, "hello"), (20, "world")],
        returning=True,
    )
    assert cur.rowcount == 1
    assert cur.statusmessage.startswith("INSERT")
    with pytest.raises(psycopg.ProgrammingError):
        cur.fetchone()
    pgresult = cur.pgresult
    assert cur.nextset()
    assert cur.rowcount == 1
    assert cur.statusmessage.startswith("INSERT")
    assert pgresult is not cur.pgresult
    assert cur.nextset() is None


def test_executemany_rowcount_no_hit(conn, execmany):
    cur = conn.cursor()
    cur.executemany(ph(cur, "delete from execmany where id = %s"), [(-1,), (-2,)])
    assert cur.rowcount == 0
    cur.executemany(ph(cur, "delete from execmany where id = %s"), [])
    assert cur.rowcount == 0
    cur.executemany(
        ph(cur, "delete from execmany where id = %s returning num"), [(-1,), (-2,)]
    )
    assert cur.rowcount == 0


@pytest.mark.parametrize(
    "query",
    [
        "insert into nosuchtable values (%s, %s)",
        "copy (select %s, %s) to stdout",
        "wat (%s, %s)",
    ],
)
def test_executemany_badquery(conn, query):
    cur = conn.cursor()
    with pytest.raises(psycopg.DatabaseError):
        cur.executemany(ph(cur, query), [(10, "hello"), (20, "world")])


@pytest.mark.parametrize("fmt_in", PyFormat)
def test_executemany_null_first(conn, fmt_in):
    cur = conn.cursor()
    cur.execute("create table testmany (a bigint, b bigint)")
    cur.executemany(
        ph(cur, f"insert into testmany values (%{fmt_in.value}, %{fmt_in.value})"),
        [[1, None], [3, 4]],
    )
    with pytest.raises((psycopg.DataError, psycopg.ProgrammingError)):
        cur.executemany(
            ph(cur, f"insert into testmany values (%{fmt_in.value}, %{fmt_in.value})"),
            [[1, ""], [3, 4]],
        )


@pytest.mark.slow
def test_executemany_lock(conn):

    def do_execmany():
        with conn.cursor() as cur:
            cur.executemany(ph(cur, "select pg_sleep(%s)"), [(0.1,) for _ in range(10)])

    def do_exec():
        with conn.cursor() as cur:
            for i in range(100):
                cur.execute("select 1")

    gather(spawn(do_execmany), spawn(do_exec))


def test_rowcount(conn):
    cur = conn.cursor()

    cur.execute("select 1 from generate_series(1, 0)")
    assert cur.rowcount == 0

    cur.execute("select 1 from generate_series(1, 42)")
    assert cur.rowcount == 42

    cur.execute("show timezone")
    assert cur.rowcount == 1

    cur.execute("create table test_rowcount_notuples (id int primary key)")
    assert cur.rowcount == -1

    cur.execute("insert into test_rowcount_notuples select generate_series(1, 42)")
    assert cur.rowcount == 42


def test_rownumber(conn):
    cur = conn.cursor()
    assert cur.rownumber is None

    cur.execute("select 1 from generate_series(1, 42)")
    assert cur.rownumber == 0

    cur.fetchone()
    assert cur.rownumber == 1
    cur.fetchone()
    assert cur.rownumber == 2
    cur.fetchmany(10)
    assert cur.rownumber == 12
    rns: list[int] = []
    for i in cur:
        assert cur.rownumber
        rns.append(cur.rownumber)
        if len(rns) >= 3:
            break
    assert rns == [13, 14, 15]
    assert len(cur.fetchall()) == 42 - rns[-1]
    assert cur.rownumber == 42


@pytest.mark.parametrize("query", ["", "set timezone to utc"])
def test_rownumber_none(conn, query):
    cur = conn.cursor()
    cur.execute(query)
    assert cur.rownumber is None


def test_rownumber_mixed(conn):
    cur = conn.cursor()
    cur.execute("""
select x from generate_series(1, 3) x;
set timezone to utc;
select x from generate_series(4, 6) x;
""")
    assert cur.rownumber == 0
    assert cur.fetchone() == (1,)
    assert cur.rownumber == 1
    assert cur.fetchone() == (2,)
    assert cur.rownumber == 2
    cur.nextset()
    assert cur.rownumber is None
    cur.nextset()
    assert cur.rownumber == 0
    assert cur.fetchone() == (4,)
    assert cur.rownumber == 1


def test_iter(conn):
    cur = conn.cursor()
    cur.execute("select generate_series(1, 3)")
    assert list(cur) == [(1,), (2,), (3,)]


def test_iter_stop(conn):
    cur = conn.cursor()
    cur.execute("select generate_series(1, 3)")
    for rec in cur:
        assert rec == (1,)
        break

    for rec in cur:
        assert rec == (2,)
        break

    assert cur.fetchone() == (3,)
    assert list(cur) == []


def test_row_factory(conn):
    cur = conn.cursor(row_factory=my_row_factory)

    cur.execute("reset search_path")
    with pytest.raises(psycopg.ProgrammingError):
        cur.fetchone()

    cur.execute("select 'foo' as bar")
    (r,) = cur.fetchone()
    assert r == "FOObar"

    cur.execute("select 'x' as x; select 'y' as y, 'z' as z")
    assert cur.fetchall() == [["Xx"]]
    assert cur.nextset()
    assert cur.fetchall() == [["Yy", "Zz"]]

    cur.scroll(-1)
    cur.row_factory = rows.dict_row
    assert cur.fetchone() == {"y": "y", "z": "z"}


def test_row_factory_none(conn):
    cur = conn.cursor(row_factory=None)
    assert cur.row_factory is rows.tuple_row
    cur.execute("select 1 as a, 2 as b")
    r = cur.fetchone()
    assert type(r) is tuple
    assert r == (1, 2)


def test_bad_row_factory(conn):

    def broken_factory(cur):
        1 / 0

    cur = conn.cursor(row_factory=broken_factory)
    with pytest.raises(ZeroDivisionError):
        cur.execute("select 1")

    def broken_maker(cur):

        def make_row(seq):
            1 / 0

        return make_row

    cur = conn.cursor(row_factory=broken_maker)
    cur.execute("select 1")
    with pytest.raises(ZeroDivisionError):
        cur.fetchone()


def test_scroll(conn):
    cur = conn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        cur.scroll(0)

    cur.execute("select generate_series(0,9)")
    cur.scroll(2)
    assert cur.fetchone() == (2,)
    cur.scroll(2)
    assert cur.fetchone() == (5,)
    cur.scroll(2, mode="relative")
    assert cur.fetchone() == (8,)
    cur.scroll(-1)
    assert cur.fetchone() == (8,)
    cur.scroll(-2)
    assert cur.fetchone() == (7,)
    cur.scroll(2, mode="absolute")
    assert cur.fetchone() == (2,)

    # on the boundary
    cur.scroll(0, mode="absolute")
    assert cur.fetchone() == (0,)
    with pytest.raises(IndexError):
        cur.scroll(-1, mode="absolute")

    cur.scroll(0, mode="absolute")
    with pytest.raises(IndexError):
        cur.scroll(-1)

    cur.scroll(9, mode="absolute")
    assert cur.fetchone() == (9,)
    with pytest.raises(IndexError):
        cur.scroll(10, mode="absolute")

    cur.scroll(9, mode="absolute")
    with pytest.raises(IndexError):
        cur.scroll(1)

    with pytest.raises(ValueError):
        cur.scroll(1, "wat")


@pytest.mark.parametrize(
    "query, params, want",
    [
        ("select %(x)s", {"x": 1}, (1,)),
        ("select %(x)s, %(y)s", {"x": 1, "y": 2}, (1, 2)),
        ("select %(x)s, %(x)s", {"x": 1}, (1, 1)),
    ],
)
def test_execute_params_named(conn, query, params, want):
    cur = conn.cursor()
    cur.execute(ph(cur, query), params)
    rec = cur.fetchone()
    assert rec == want


def test_stream(conn):
    cur = conn.cursor()
    recs = []
    for rec in cur.stream(
        ph(cur, "select i, '2021-01-01'::date + i from generate_series(1, %s) as i"),
        [2],
    ):
        recs.append(rec)

    assert recs == [(1, dt.date(2021, 1, 2)), (2, dt.date(2021, 1, 3))]


def test_stream_sql(conn):
    cur = conn.cursor()
    recs = list(
        cur.stream(
            sql.SQL(
                "select i, '2021-01-01'::date + i from generate_series(1, {}) as i"
            ).format(2)
        )
    )

    assert recs == [(1, dt.date(2021, 1, 2)), (2, dt.date(2021, 1, 3))]


def test_stream_row_factory(conn):
    cur = conn.cursor(row_factory=rows.dict_row)
    it = cur.stream("select generate_series(1,2) as a")
    assert next(it)["a"] == 1
    cur.row_factory = rows.namedtuple_row
    assert next(it).a == 2


def test_stream_no_row(conn):
    cur = conn.cursor()
    recs = list(cur.stream("select generate_series(2,1) as a"))
    assert recs == []


def test_stream_chunked_invalid_size(conn):
    cur = conn.cursor()
    with pytest.raises(ValueError, match="size must be >= 1"):
        next(cur.stream("select 1", size=0))


@pytest.mark.libpq("< 17")
def test_stream_chunked_not_supported(conn):
    cur = conn.cursor()
    with pytest.raises(psycopg.NotSupportedError):
        next(cur.stream("select generate_series(1, 4)", size=2))


@pytest.mark.libpq(">= 17")
def test_stream_chunked(conn):
    cur = conn.cursor()
    recs = list(cur.stream("select generate_series(1, 5) as a", size=2))
    assert recs == [(1,), (2,), (3,), (4,), (5,)]


@pytest.mark.libpq(">= 17")
def test_stream_chunked_row_factory(conn):
    cur = conn.cursor(row_factory=rows.scalar_row)
    it = cur.stream("select generate_series(1, 5) as a", size=2)
    for i in range(1, 6):
        assert next(it) == i
        assert [c.name for c in cur.description] == ["a"]


@pytest.mark.crdb_skip("no col query")
def test_stream_no_col(conn):
    cur = conn.cursor()
    recs = list(cur.stream("select"))
    assert recs == [()]


@pytest.mark.parametrize(
    "query", ["create table test_stream_badq ()", "copy (select 1) to stdout", "wat?"]
)
def test_stream_badquery(conn, query):
    cur = conn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        for rec in cur.stream(query):
            pass


def test_stream_error_tx(conn):
    cur = conn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        for rec in cur.stream("wat"):
            pass
    assert conn.info.transaction_status == pq.TransactionStatus.INERROR


def test_stream_error_notx(conn):
    conn.set_autocommit(True)
    cur = conn.cursor()
    with pytest.raises(psycopg.ProgrammingError):
        for rec in cur.stream("wat"):
            pass
    assert conn.info.transaction_status == pq.TransactionStatus.IDLE


def test_stream_error_python_to_consume(conn):
    cur = conn.cursor()
    with pytest.raises(ZeroDivisionError):
        with closing(cur.stream("select generate_series(1, 10000)")) as gen:
            for rec in gen:
                1 / 0
    assert conn.info.transaction_status in (
        pq.TransactionStatus.INTRANS,
        pq.TransactionStatus.INERROR,
    )


def test_stream_error_python_consumed(conn):
    cur = conn.cursor()
    with pytest.raises(ZeroDivisionError):
        gen = cur.stream("select 1")
        for rec in gen:
            1 / 0

    gen.close()
    assert conn.info.transaction_status == pq.TransactionStatus.INTRANS


@pytest.mark.parametrize("autocommit", [False, True])
def test_stream_close(conn, autocommit):
    conn.set_autocommit(autocommit)
    cur = conn.cursor()
    with pytest.raises(psycopg.OperationalError):
        for rec in cur.stream("select generate_series(1, 3)"):
            if rec[0] == 1:
                conn.close()
            else:
                assert False

    assert conn.closed


def test_stream_binary_cursor(conn):
    with raiseif(
        conn.cursor_factory is psycopg.ClientCursor, psycopg.NotSupportedError
    ):
        cur = conn.cursor(binary=True)
        recs = []
        for rec in cur.stream("select x::int4 from generate_series(1, 2) x"):
            recs.append(rec)
            assert cur.pgresult.fformat(0) == 1
            assert cur.pgresult.get_value(0, 0) == bytes([0, 0, 0, rec[0]])

        assert recs == [(1,), (2,)]


def test_stream_execute_binary(conn):
    cur = conn.cursor()
    recs = []
    with raiseif(
        conn.cursor_factory is psycopg.ClientCursor, psycopg.NotSupportedError
    ):
        for rec in cur.stream(
            "select x::int4 from generate_series(1, 2) x", binary=True
        ):
            recs.append(rec)
            assert cur.pgresult.fformat(0) == 1
            assert cur.pgresult.get_value(0, 0) == bytes([0, 0, 0, rec[0]])

        assert recs == [(1,), (2,)]


def test_stream_binary_cursor_text_override(conn):
    cur = conn.cursor(binary=True)
    recs = []
    for rec in cur.stream("select generate_series(1, 2)", binary=False):
        recs.append(rec)
        assert cur.pgresult.fformat(0) == 0
        assert cur.pgresult.get_value(0, 0) == str(rec[0]).encode()

    assert recs == [(1,), (2,)]


def test_str(conn):
    cur = conn.cursor()
    assert "[IDLE]" in str(cur)
    assert "[closed]" not in str(cur)
    assert "[no result]" in str(cur)
    cur.execute("select 1")
    assert "[INTRANS]" in str(cur)
    assert "[TUPLES_OK]" in str(cur)
    assert "[closed]" not in str(cur)
    assert "[no result]" not in str(cur)
    cur.close()
    assert "[closed]" in str(cur)
    assert "[INTRANS]" in str(cur)


@pytest.mark.pipeline
def test_message_0x33(conn):
    # https://github.com/psycopg/psycopg/issues/314
    notices = []
    conn.add_notice_handler(lambda diag: notices.append(diag.message_primary))

    conn.set_autocommit(True)
    with conn.pipeline():
        cur = conn.execute("select 'test'")
        assert cur.fetchone() == ("test",)

    assert not notices


def test_typeinfo(conn):
    info = TypeInfo.fetch(conn, "jsonb")
    assert info is not None


def test_error_no_result(conn):
    cur = conn.cursor()
    with pytest.raises(psycopg.ProgrammingError, match="no result available"):
        cur.fetchone()

    cur.execute("set timezone to utc")
    with pytest.raises(
        psycopg.ProgrammingError, match="last operation.*command status: SET"
    ):
        cur.fetchone()

    cur.execute("")
    with pytest.raises(
        psycopg.ProgrammingError, match="last operation.*result status: EMPTY_QUERY"
    ):
        cur.fetchone()


def test_row_maker_returns_none(conn):
    cur = conn.cursor(row_factory=rows.scalar_row)
    query = "values (null), (0)"
    recs = [None, 0]

    cur.execute(query)
    assert [cur.fetchone() for _ in range(len(recs))] == recs
    cur.execute(query)
    assert cur.fetchmany(len(recs)) == recs
    cur.execute(query)
    assert cur.fetchall() == recs
    cur.execute(query)
    assert list(cur) == recs
    stream = cur.stream(query)
    assert list(stream) == recs


@pytest.mark.parametrize("count", [1, 3])
def test_results_after_execute(conn, count):
    with conn.cursor() as cur:
        cur.execute(
            ";".join((f"select * from generate_series(1, {i})" for i in range(count)))
        )
        ress = list((res.fetchall() for res in cur.results()))
        assert ress == [[(j + 1,) for j in range(i)] for i in range(count)]


@pytest.mark.parametrize("count", [0, 1, 3])
@pytest.mark.parametrize("returning", [False, True])
def test_results_after_executemany(conn, count, returning):
    with conn.cursor() as cur:
        cur.executemany(
            ph(cur, "select * from generate_series(1, %s)"),
            [(i,) for i in range(count)],
            returning=returning,
        )
        ress = list((res.fetchall() for res in cur.results()))
        if returning:
            assert ress == [[(j + 1,) for j in range(i)] for i in range(count)]
        else:
            assert ress == []


def test_change_loader_results(conn):
    cur = conn.cursor()
    # With no result
    cur.adapters.register_loader("text", make_loader("1"))

    cur.execute("""
        values ('foo'::text);
        values ('bar'::text), ('baz');
        values ('qux'::text);
        """)
    assert cur.fetchall() == [("foo1",)]

    cur.nextset()
    assert cur.fetchone() == ("bar1",)
    cur.adapters.register_loader("text", make_loader("2"))
    assert cur.fetchone() == ("baz2",)
    cur.scroll(-2)
    assert cur.fetchall() == [("bar2",), ("baz2",)]

    cur.nextset()
    assert cur.fetchall() == [("qux2",)]

    # After the final result
    assert not cur.nextset()
    cur.adapters.register_loader("text", make_loader("3"))
    assert cur.fetchone() is None
    cur.set_result(0)
    assert cur.fetchall() == [("foo3",)]



# test_sql.py - tests for the psycopg2.sql module

# Copyright (C) 2020 The Psycopg Team

import re
import datetime as dt

import pytest

from psycopg import ProgrammingError, pq, sql
from psycopg.adapt import PyFormat
from psycopg.types import TypeInfo
from psycopg._encodings import py2pgenc
from psycopg.types.string import StrDumper

from .utils import eur
from .fix_crdb import crdb_encoding, crdb_scs_off
from .test_adapt import make_dumper


@pytest.mark.parametrize(
    "obj, quoted",
    [
        ("foo\\bar", " E'foo\\\\bar'"),
        ("hello", "'hello'"),
        (42, "42"),
        (True, "true"),
        (None, "NULL"),
    ],
)
def test_quote(obj, quoted):
    assert sql.quote(obj) == quoted


@pytest.mark.parametrize("scs", ["on", crdb_scs_off("off")])
def test_quote_roundtrip(conn, scs):
    messages = []
    conn.add_notice_handler(lambda msg: messages.append(msg.message_primary))
    conn.execute(f"set standard_conforming_strings to {scs}")

    for i in range(1, 256):
        want = chr(i)
        quoted = sql.quote(want)
        got = conn.execute(f"select {quoted}::text").fetchone()[0]
        assert want == got

        # No "nonstandard use of \\ in a string literal" warning
        assert not messages, f"error with {want!r}"


@pytest.mark.parametrize("dummy", [crdb_scs_off("off")])
def test_quote_stable_despite_deranged_libpq(conn, dummy):
    # Verify the libpq behaviour of PQescapeString using the last setting seen.
    # Check that we are not affected by it.
    good_str = " E'\\\\'"
    good_bytes = " E'\\\\000'::bytea"
    conn.execute("set standard_conforming_strings to on")
    assert pq.Escaping().escape_string(b"\\") == b"\\"
    assert sql.quote("\\") == good_str
    assert pq.Escaping().escape_bytea(b"\x00") == b"\\000"
    assert sql.quote(b"\x00") == good_bytes

    conn.execute("set standard_conforming_strings to off")
    assert pq.Escaping().escape_string(b"\\") == b"\\\\"
    assert sql.quote("\\") == good_str
    assert pq.Escaping().escape_bytea(b"\x00") == b"\\\\000"
    assert sql.quote(b"\x00") == good_bytes

    # Verify that the good values are actually good
    messages = []
    conn.add_notice_handler(lambda msg: messages.append(msg.message_primary))
    conn.execute("set escape_string_warning to on")
    for scs in ("on", "off"):
        conn.execute(f"set standard_conforming_strings to {scs}")
        cur = conn.execute(f"select {good_str}, {good_bytes}::bytea")
        assert cur.fetchone() == ("\\", b"\x00")

    # No "nonstandard use of \\ in a string literal" warning
    assert not messages


class TestSqlFormat:
    def test_pos(self, conn):
        s = sql.SQL("select {} from {}").format(
            sql.Identifier("field"), sql.Identifier("table")
        )
        s1 = s.as_string(conn)
        assert isinstance(s1, str)
        assert s1 == 'select "field" from "table"'

    def test_pos_spec(self, conn):
        s = sql.SQL("select {0} from {1}").format(
            sql.Identifier("field"), sql.Identifier("table")
        )
        s1 = s.as_string(conn)
        assert isinstance(s1, str)
        assert s1 == 'select "field" from "table"'

        s = sql.SQL("select {1} from {0}").format(
            sql.Identifier("table"), sql.Identifier("field")
        )
        s1 = s.as_string(conn)
        assert isinstance(s1, str)
        assert s1 == 'select "field" from "table"'

    def test_dict(self, conn):
        s = sql.SQL("select {f} from {t}").format(
            f=sql.Identifier("field"), t=sql.Identifier("table")
        )
        s1 = s.as_string(conn)
        assert isinstance(s1, str)
        assert s1 == 'select "field" from "table"'

    def test_compose_literal(self, conn):
        s = sql.SQL("select {0};").format(sql.Literal(dt.date(2016, 12, 31)))
        s1 = s.as_string(conn)
        assert s1 == "select '2016-12-31'::date;"

    def test_compose_empty(self, conn):
        s = sql.SQL("select foo;").format()
        s1 = s.as_string(conn)
        assert s1 == "select foo;"

    def test_percent_escape(self, conn):
        s = sql.SQL("42 % {0}").format(sql.Literal(7))
        s1 = s.as_string(conn)
        assert s1 == "42 % 7"

    def test_braces_escape(self, conn):
        s = sql.SQL("{{{0}}}").format(sql.Literal(7))
        assert s.as_string(conn) == "{7}"
        s = sql.SQL("{{1,{0}}}").format(sql.Literal(7))
        assert s.as_string(conn) == "{1,7}"

    def test_compose_badnargs(self):
        with pytest.raises(IndexError):
            sql.SQL("select {0};").format()

    def test_compose_badnargs_auto(self):
        with pytest.raises(IndexError):
            sql.SQL("select {};").format()
        with pytest.raises(ValueError):
            sql.SQL("select {} {1};").format(10, 20)
        with pytest.raises(ValueError):
            sql.SQL("select {0} {};").format(10, 20)

    def test_compose_bad_args_type(self):
        with pytest.raises(IndexError):
            sql.SQL("select {0};").format(a=10)
        with pytest.raises(KeyError):
            sql.SQL("select {x};").format(10)

    def test_no_modifiers(self):
        with pytest.raises(ValueError):
            sql.SQL("select {a!r};").format(a=10)
        with pytest.raises(ValueError):
            sql.SQL("select {a:<};").format(a=10)

    def test_must_be_adaptable(self, conn):
        class Foo:
            pass

        s = sql.SQL("select {0};").format(sql.Literal(Foo()))
        with pytest.raises(ProgrammingError):
            s.as_string(conn)

    def test_auto_literal(self, conn):
        s = sql.SQL("select {}, {}, {}").format("he'lo", 10, dt.date(2020, 1, 1))
        assert s.as_string(conn) == "select 'he''lo', 10, '2020-01-01'::date"

    def test_execute(self, conn):
        cur = conn.cursor()
        cur.execute("""
            create table test_compose (
                id serial primary key,
                foo text, bar text, "ba'z" text)
            """)
        cur.execute(
            sql.SQL("insert into {0} (id, {1}) values (%s, {2})").format(
                sql.Identifier("test_compose"),
                sql.SQL(", ").join(map(sql.Identifier, ["foo", "bar", "ba'z"])),
                (sql.Placeholder() * 3).join(", "),
            ),
            (10, "a", "b", "c"),
        )

        cur.execute("select * from test_compose")
        assert cur.fetchall() == [(10, "a", "b", "c")]

    def test_executemany(self, conn):
        cur = conn.cursor()
        cur.execute("""
            create table test_compose (
                id serial primary key,
                foo text, bar text, "ba'z" text)
            """)
        cur.executemany(
            sql.SQL("insert into {0} (id, {1}) values (%s, {2})").format(
                sql.Identifier("test_compose"),
                sql.SQL(", ").join(map(sql.Identifier, ["foo", "bar", "ba'z"])),
                (sql.Placeholder() * 3).join(", "),
            ),
            [(10, "a", "b", "c"), (20, "d", "e", "f")],
        )

        cur.execute("select * from test_compose")
        assert cur.fetchall() == [(10, "a", "b", "c"), (20, "d", "e", "f")]

    @pytest.mark.crdb_skip("copy")
    def test_copy(self, conn):
        cur = conn.cursor()
        cur.execute("""
            create table test_compose (
                id serial primary key,
                foo text, bar text, "ba'z" text)
            """)

        with cur.copy(
            sql.SQL("copy {t} (id, foo, bar, {f}) from stdin").format(
                t=sql.Identifier("test_compose"), f=sql.Identifier("ba'z")
            ),
        ) as copy:
            copy.write_row((10, "a", "b", "c"))
            copy.write_row((20, "d", "e", "f"))

        with cur.copy(
            sql.SQL("copy (select {f} from {t} order by id) to stdout").format(
                t=sql.Identifier("test_compose"), f=sql.Identifier("ba'z")
            )
        ) as copy:
            assert list(copy) == [b"c\n", b"f\n"]


class TestIdentifier:
    def test_class(self):
        assert issubclass(sql.Identifier, sql.Composable)

    def test_init(self):
        assert isinstance(sql.Identifier("foo"), sql.Identifier)
        assert isinstance(sql.Identifier("foo"), sql.Identifier)
        assert isinstance(sql.Identifier("foo", "bar", "baz"), sql.Identifier)
        with pytest.raises(TypeError):
            sql.Identifier()
        with pytest.raises(TypeError):
            sql.Identifier(10)  # type: ignore[arg-type]
        with pytest.raises(TypeError):
            sql.Identifier(dt.date(2016, 12, 31))  # type: ignore[arg-type]

    def test_repr(self):
        obj = sql.Identifier("fo'o")
        assert repr(obj) == 'Identifier("fo\'o")'
        assert repr(obj) == str(obj)

        obj = sql.Identifier("fo'o", 'ba"r')
        assert repr(obj) == "Identifier(\"fo'o\", 'ba\"r')"
        assert repr(obj) == str(obj)

    def test_eq(self):
        assert sql.Identifier("foo") == sql.Identifier("foo")
        assert sql.Identifier("foo", "bar") == sql.Identifier("foo", "bar")
        assert sql.Identifier("foo") != sql.Identifier("bar")
        assert sql.Identifier("foo") != "foo"
        assert sql.Identifier("foo") != sql.SQL("foo")

    _as_string_params = [
        (("foo",), '"foo"'),
        (("foo", "bar"), '"foo"."bar"'),
        (("fo'o", 'ba"r'), '"fo\'o"."ba""r"'),
    ]

    @pytest.mark.parametrize("args, want", _as_string_params)
    def test_as_string(self, conn, args, want):
        assert sql.Identifier(*args).as_string(conn) == want

    @pytest.mark.parametrize("args, want", _as_string_params)
    def test_as_string_no_conn(self, args, want):
        assert sql.Identifier(*args).as_string(None) == want
        assert sql.Identifier(*args).as_string() == want

    _as_bytes_params = [
        crdb_encoding(("foo",), '"foo"', "ascii"),
        crdb_encoding(("foo", "bar"), '"foo"."bar"', "ascii"),
        crdb_encoding(("fo'o", 'ba"r'), '"fo\'o"."ba""r"', "ascii"),
        (("foo", eur), f'"foo"."{eur}"', "utf8"),
        crdb_encoding(("foo", eur), f'"foo"."{eur}"', "latin9"),
    ]

    @pytest.mark.parametrize("args, want, enc", _as_bytes_params)
    def test_as_bytes(self, conn, args, want, enc):
        want = want.encode(enc)
        conn.execute(f"set client_encoding to {py2pgenc(enc).decode()}")
        got = sql.Identifier(*args).as_bytes(conn)
        assert isinstance(got, bytes)
        assert got == want

    @pytest.mark.parametrize("args, want, enc", _as_bytes_params)
    def test_as_bytes_no_conn(self, conn, args, want, enc):
        want = want.encode()
        assert sql.Identifier(*args).as_bytes(None) == want
        assert sql.Identifier(*args).as_bytes() == want

    def test_join(self):
        assert not hasattr(sql.Identifier("foo"), "join")

    def test_escape_no_conn(self, conn):
        conn.execute("set client_encoding to 'utf8'")
        for c in range(1, 128):
            s = chr(c)
            want = sql.Identifier(s).as_bytes(conn)
            assert want == sql.Identifier(s).as_bytes(None)


class TestLiteral:
    def test_class(self):
        assert issubclass(sql.Literal, sql.Composable)

    def test_init(self):
        assert isinstance(sql.Literal("foo"), sql.Literal)
        assert isinstance(sql.Literal("foo"), sql.Literal)
        assert isinstance(sql.Literal(b"foo"), sql.Literal)
        assert isinstance(sql.Literal(42), sql.Literal)
        assert isinstance(sql.Literal(dt.date(2016, 12, 31)), sql.Literal)

    def test_repr(self):
        assert repr(sql.Literal("foo")) == "Literal('foo')"
        assert str(sql.Literal("foo")) == "Literal('foo')"

    _params = [
        (None, "NULL"),
        ("foo", "'foo'"),
        (42, "42"),
        (dt.date(2017, 1, 1), "'2017-01-01'::date"),
    ]

    @pytest.mark.parametrize("obj, want", _params)
    def test_as_string(self, conn, obj, want):
        got = sql.Literal(obj).as_string(conn)
        if isinstance(obj, str):
            got = no_e(got)
        assert got == want

    @pytest.mark.parametrize("obj, want", _params)
    def test_as_bytes(self, conn, obj, want):
        got = sql.Literal(obj).as_bytes(conn)
        assert isinstance(got, bytes)
        if isinstance(obj, str):
            got = no_e(got)
        assert got == want.encode()

    @pytest.mark.parametrize("obj, want", _params)
    def test_as_string_no_conn(self, obj, want):
        got = sql.Literal(obj).as_string()
        if isinstance(obj, str):
            got = no_e(got)
        assert got == want

    @pytest.mark.parametrize("obj, want", _params)
    def test_as_bytes_no_conn(self, obj, want):
        got = sql.Literal(obj).as_bytes()
        if isinstance(obj, str):
            got = no_e(got)
        assert got == want.encode()

    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
    def test_as_bytes_encoding(self, conn, encoding):
        conn.execute(f"set client_encoding to {encoding}")
        assert sql.Literal(eur).as_bytes(conn) == f"'{eur}'".encode(encoding)

    def test_eq(self):
        assert sql.Literal("foo") == sql.Literal("foo")
        assert sql.Literal("foo") != sql.Literal("bar")
        assert sql.Literal("foo") != "foo"
        assert sql.Literal("foo") != sql.SQL("foo")

    def test_must_be_adaptable(self, conn):
        class Foo:
            pass

        with pytest.raises(ProgrammingError):
            sql.Literal(Foo()).as_string(conn)

    def test_array(self, conn):
        assert (
            sql.Literal([dt.date(2000, 1, 1)]).as_string(conn)
            == "'{2000-01-01}'::date[]"
        )

    def test_short_name_builtin(self, conn):
        assert sql.Literal(dt.time(0, 0)).as_string(conn) == "'00:00:00'::time"
        assert (
            sql.Literal(dt.datetime(2000, 1, 1)).as_string(conn)
            == "'2000-01-01 00:00:00'::timestamp"
        )
        assert (
            sql.Literal([dt.datetime(2000, 1, 1)]).as_string(conn)
            == "'{\"2000-01-01 00:00:00\"}'::timestamp[]"
        )

    def test_text_literal(self, conn):
        conn.adapters.register_dumper(str, StrDumper)
        assert sql.Literal("foo").as_string(conn) == "'foo'"

    @pytest.mark.crdb_skip("composite")  # create type, actually
    @pytest.mark.parametrize("name", ["a-b", f"{eur}", "order", "foo bar", "FooBar"])
    def test_invalid_name(self, conn, name):
        if conn.info.parameter_status("is_superuser") != "on":
            pytest.skip("not a superuser")
        conn.execute(f"""
            set client_encoding to utf8;
            create type "{name}";
            create function invin(cstring) returns "{name}"
                language internal immutable strict as 'textin';
            create function invout("{name}") returns cstring
                language internal immutable strict as 'textout';
            create type "{name}" (input=invin, output=invout, like=text);
            """)
        info = TypeInfo.fetch(conn, f'"{name}"')

        class InvDumper(StrDumper):
            oid = info.oid

            def dump(self, obj):
                rv = super().dump(obj)
                return b"%s-inv" % rv

        info.register(conn)
        conn.adapters.register_dumper(str, InvDumper)

        assert sql.Literal("hello").as_string(conn) == f"'hello-inv'::\"{name}\""
        cur = conn.execute(sql.SQL("select {}").format("hello"))
        assert cur.fetchone()[0] == "hello-inv"

        assert (
            sql.Literal(["hello"]).as_string(conn) == f"'{{hello-inv}}'::\"{name}\"[]"
        )
        cur = conn.execute(sql.SQL("select {}").format(["hello"]))
        assert cur.fetchone()[0] == ["hello-inv"]


class TestSQL:
    def test_class(self):
        assert issubclass(sql.SQL, sql.Composable)

    def test_init(self):
        assert isinstance(sql.SQL("foo"), sql.SQL)
        assert isinstance(sql.SQL("foo"), sql.SQL)
        with pytest.raises(TypeError):
            sql.SQL(10)  # type: ignore[arg-type]
        with pytest.raises(TypeError):
            sql.SQL(dt.date(2016, 12, 31))  # type: ignore[arg-type]

    def test_repr(self, conn):
        assert repr(sql.SQL("foo")) == "SQL('foo')"
        assert str(sql.SQL("foo")) == "SQL('foo')"
        assert sql.SQL("foo").as_string(conn) == "foo"

    def test_eq(self):
        assert sql.SQL("foo") == sql.SQL("foo")
        assert sql.SQL("foo") != sql.SQL("bar")
        assert sql.SQL("foo") != "foo"
        assert sql.SQL("foo") != sql.Literal("foo")

    def test_sum(self, conn):
        obj = sql.SQL("foo") + sql.SQL("bar")
        assert isinstance(obj, sql.Composed)
        assert obj.as_string(conn) == "foobar"

    def test_sum_inplace(self, conn):
        obj = sql.SQL("f") + sql.SQL("oo")
        obj += sql.SQL("bar")
        assert isinstance(obj, sql.Composed)
        assert obj.as_string(conn) == "foobar"

    def test_multiply(self, conn):
        obj = sql.SQL("foo") * 3
        assert isinstance(obj, sql.Composed)
        assert obj.as_string(conn) == "foofoofoo"

    def test_join(self, conn):
        obj = sql.SQL(", ").join(
            [sql.Identifier("foo"), sql.SQL("bar"), sql.Literal(42)]
        )
        assert isinstance(obj, sql.Composed)
        assert obj.as_string(conn) == '"foo", bar, 42'

        obj = sql.SQL(", ").join(
            sql.Composed([sql.Identifier("foo"), sql.SQL("bar"), sql.Literal(42)])
        )
        assert isinstance(obj, sql.Composed)
        assert obj.as_string(conn) == '"foo", bar, 42'

        obj = sql.SQL(", ").join(["foo", "bar", 42])
        assert isinstance(obj, sql.Composed)
        assert obj.as_string(conn) == """'foo', 'bar', 42"""

        obj2 = sql.SQL(", ").join([])
        assert obj2 == sql.Composed([])

    def test_as_string(self, conn):
        assert sql.SQL("foo").as_string(conn) == "foo"
        assert sql.SQL("foo").as_string() == "foo"

    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
    def test_as_bytes(self, conn, encoding):
        if encoding:
            conn.execute(f"set client_encoding to {encoding}")

        got = sql.SQL(eur).as_bytes(conn)
        assert isinstance(got, bytes)
        assert got == eur.encode(encoding)

    def test_no_conn(self):
        assert sql.SQL(eur).as_string() == eur
        assert sql.SQL(eur).as_bytes() == eur.encode()


class TestComposed:
    def test_class(self):
        assert issubclass(sql.Composed, sql.Composable)

    def test_repr(self):
        obj = sql.Composed([sql.Literal("foo"), sql.Identifier("b'ar")])
        assert repr(obj) == """Composed([Literal('foo'), Identifier("b'ar")])"""
        assert str(obj) == repr(obj)

    def test_eq(self):
        L = [sql.Literal("foo"), sql.Identifier("b'ar")]
        l2 = [sql.Literal("foo"), sql.Literal("b'ar")]
        assert sql.Composed(L) == sql.Composed(list(L))
        assert sql.Composed(L) != L
        assert sql.Composed(L) != sql.Composed(l2)

    def test_join(self, conn):
        obj = sql.Composed([sql.Literal("foo"), sql.Identifier("b'ar")])
        obj = obj.join(", ")
        assert isinstance(obj, sql.Composed)
        assert no_e(obj.as_string(conn)) == "'foo', \"b'ar\""

    def test_auto_literal(self, conn):
        obj = sql.Composed(["fo'o", dt.date(2020, 1, 1)])
        obj = obj.join(", ")
        assert isinstance(obj, sql.Composed)
        assert no_e(obj.as_string(conn)) == "'fo''o', '2020-01-01'::date"

    def test_sum(self, conn):
        obj = sql.Composed([sql.SQL("foo ")])
        obj = obj + sql.Literal("bar")
        assert isinstance(obj, sql.Composed)
        assert no_e(obj.as_string(conn)) == "foo 'bar'"

    def test_sum_inplace(self, conn):
        obj = sql.Composed([sql.SQL("foo ")])
        obj += sql.Literal("bar")
        assert isinstance(obj, sql.Composed)
        assert no_e(obj.as_string(conn)) == "foo 'bar'"

        obj = sql.Composed([sql.SQL("foo ")])
        obj += sql.Composed([sql.Literal("bar")])
        assert isinstance(obj, sql.Composed)
        assert no_e(obj.as_string(conn)) == "foo 'bar'"

    def test_iter(self):
        obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
        it = iter(obj)
        i = next(it)
        assert i == sql.SQL("foo")
        i = next(it)
        assert i == sql.SQL("bar")
        with pytest.raises(StopIteration):
            next(it)

    def test_as_string(self, conn):
        obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
        assert obj.as_string(conn) == "foobar"
        assert obj.as_string() == "foobar"

    def test_as_bytes(self, conn):
        obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
        assert obj.as_bytes(conn) == b"foobar"
        assert obj.as_bytes() == b"foobar"

    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
    def test_as_bytes_encoding(self, conn, encoding):
        obj = sql.Composed([sql.SQL("foo"), sql.SQL(eur)])
        conn.execute(f"set client_encoding to {encoding}")
        assert obj.as_bytes(conn) == ("foo" + eur).encode(encoding)


class TestPlaceholder:
    def test_class(self):
        assert issubclass(sql.Placeholder, sql.Composable)

    @pytest.mark.parametrize("format", PyFormat)
    def test_repr_format(self, conn, format):
        ph = sql.Placeholder(format=format)
        add = f"format={format.name}" if format != PyFormat.AUTO else ""
        assert str(ph) == repr(ph) == f"Placeholder({add})"

    @pytest.mark.parametrize("format", PyFormat)
    def test_repr_name_format(self, conn, format):
        ph = sql.Placeholder("foo", format=format)
        add = f", format={format.name}" if format != PyFormat.AUTO else ""
        assert str(ph) == repr(ph) == f"Placeholder('foo'{add})"

    def test_bad_name(self):
        with pytest.raises(ValueError):
            sql.Placeholder(")")

    def test_eq(self):
        assert sql.Placeholder("foo") == sql.Placeholder("foo")
        assert sql.Placeholder("foo") != sql.Placeholder("bar")
        assert sql.Placeholder("foo") != "foo"
        assert sql.Placeholder() == sql.Placeholder()
        assert sql.Placeholder("foo") != sql.Placeholder()
        assert sql.Placeholder("foo") != sql.Literal("foo")

    @pytest.mark.parametrize("format", PyFormat)
    def test_as_string(self, conn, format):
        ph = sql.Placeholder(format=format)
        assert ph.as_string(conn) == f"%{format.value}"
        assert ph.as_string() == f"%{format.value}"

        ph = sql.Placeholder(name="foo", format=format)
        assert ph.as_string(conn) == f"%(foo){format.value}"
        assert ph.as_string() == f"%(foo){format.value}"

    @pytest.mark.parametrize("format", PyFormat)
    def test_as_bytes(self, conn, format):
        ph = sql.Placeholder(format=format)
        assert isinstance(ph.as_bytes(), bytes)
        assert ph.as_bytes(conn) == f"%{format.value}".encode()
        assert ph.as_bytes() == f"%{format.value}".encode()

        ph = sql.Placeholder(name="foo", format=format)
        assert ph.as_bytes(conn) == f"%(foo){format.value}".encode()
        assert ph.as_bytes() == f"%(foo){format.value}".encode()


def test_as_string():
    query = sql.as_string(sql.SQL("select {}").format("foo"))
    assert query == no_e("select 'foo'")


def test_as_string_context(conn):
    conn.adapters.register_dumper(str, make_dumper("1"))
    query = sql.as_string(sql.SQL("select {}").format("foo"), context=conn)
    assert query == no_e("select 'foo1'")


def test_as_string_literal():
    got = sql.as_string("query")
    assert got == no_e("'query'")

    got = sql.as_string(dt.date(1970, 1, 1))
    assert got == no_e("'1970-01-01'::date")


def test_as_bytes():
    query = sql.as_bytes(sql.SQL("select {}").format("foo"))
    assert isinstance(query, bytes)
    assert query == no_e(b"select 'foo'")


def test_as_bytes_context(conn):
    conn.adapters.register_dumper(str, make_dumper("1"))
    query = sql.as_bytes(sql.SQL("select {}").format("foo"), context=conn)
    assert isinstance(query, bytes)
    assert query == no_e(b"select 'foo1'")


def test_as_bytes_literal():
    got = sql.as_bytes("query")
    assert isinstance(got, bytes)
    assert got == no_e(b"'query'")

    got = sql.as_bytes(dt.date(1970, 1, 1))
    assert isinstance(got, bytes)
    assert got == no_e(b"'1970-01-01'::date")


class TestValues:
    def test_null(self, conn):
        assert isinstance(sql.NULL, sql.SQL)
        assert sql.NULL.as_string(conn) == "NULL"

    def test_default(self, conn):
        assert isinstance(sql.DEFAULT, sql.SQL)
        assert sql.DEFAULT.as_string(conn) == "DEFAULT"


def no_e(s):
    """Drop an eventual E from E'' quotes"""
    if isinstance(s, (memoryview, bytearray)):
        s = bytes(s)

    if isinstance(s, str):
        return re.sub(r"\bE'", "'", s)
    elif isinstance(s, bytes):
        return re.sub(rb"\bE'", b"'", s)
    else:
        raise TypeError(f"not dealing with {type(s).__name__}: {s}")
